;;
;; (c) 1997-2001 by Xref-Tech
;;

(provide 'xref)

;; try to set it to 't' if you have some troubles, then send me the
;; output (Marian).
(defvar xref-debug-mode nil)


;; if it does not work on your (MS) system, you may check this first.
(if (or (string-match "-nt" system-configuration) 
		(string-match "-win" system-configuration))
	(progn
	  (defvar xref-platform 'windows)
	  (defvar xref-slash ?\\)
	  (defvar xref-communication-file (concat (getenv "TEMP") "/xrefcomm.tmp"))
	  (defvar xref-run-batch-file (concat (getenv "TEMP") "/xrefrun.bat"))
	  (if (getenv "HOME")
		  (defvar xref-options-file (concat (getenv "HOME") "/_xrefrc"))
		(defvar xref-options-file "c:/_xrefrc")
		)
	  (defvar xref-user-identification "user")
	  )
  (defvar xref-platform 'unix)
  (defvar xref-slash ?/)
  (defvar xref-communication-file (concat "/tmp/" (getenv "LOGNAME") "x.xrf"))
  (defvar xref-run-batch-file (concat "/tmp/" (getenv "LOGNAME") "xr.sh"))
  (defvar xref-options-file (concat (getenv "HOME") "/.xrefrc"))
  (defvar xref-user-identification (concat (getenv "LOGNAME") "u"))
)



(defvar xrefsrv-process nil)
(defvar xrefxref-process nil)
(defvar ref-counter 0)
(defvar lastref-file " ")
(defvar lastref-line 0)
(defvar lastref-coll 0)
(defvar filter-initialized nil)
(defvar xref-pilot-char ?\ )
(defvar xref-last-message-filtered 0)
(defvar xref-sync-last-message "Xref2end3of1synchronised7output0")
(defvar xref-list-buffer-content nil)
(defvar xrefsrv-buffer) 
(defvar xref-move-refid) 
(defvar xref-move-usage)
(defvar xref-list-invocation-window) 
(defvar xref-pass-identifier-regexp "\\(\\'\\|\\`\\|[^A-Za-z0-9_]\\)")
(defvar newline-char (string-to-char "
"))
(defvar xref-menu-selection-prefix-lines 2)
(defvar xref-fontification nil)
(defvar xref-pushrefs-action nil)
(defvar xref-project-list)
(defvar xref-list-filter-level 0)
(defvar xref-sym-menu-filter-level 2)
(defvar xref-next-push-manual nil)
(defvar xref-last-message nil)
(defvar xref-last-referencing-action nil)
(defvar xref-directory-dep-prj-name "-- Automatic (directory dependent) --")
(defvar xref-list-buffer-message "'+' filter+;  '-' filter-;  button2(' ') inspect;  button3(<CR>) escape;")

(defvar xref-query-replace-map (make-sparse-keymap "Xref-query-replace"))
(define-key xref-query-replace-map "y" 'act)
(define-key xref-query-replace-map "Y" 'act)
(define-key xref-query-replace-map "n" 'noact)
(define-key xref-query-replace-map "N" 'noact)
(define-key xref-query-replace-map "c" 'confirm)
(define-key xref-query-replace-map "C" 'confirm)



(defvar xref-keywords-regexp
"\\b\\(auto\\|break\\|c\\(ase\\|har\\|on\\(st\\|tinue\\)\\)\\|d\\(efault\\|o\\(uble\\)?\\)\\|e\\(lse\\|num\\|xtern\\)\\|f\\(loat\\|or\\)\\|goto\\|i\\(f\\|nt\\)\\|long\\|re\\(gister\\|turn\\)\\|s\\(hort\\|i\\(gned\\|zeof\\)\\|t\\(atic\\|ruct\\)\\|witch\\)\\|typedef\\|un\\(ion\\|signed\\)\\|vo\\(id\\|latile\\)\\|while\\)\\b"
)

;; somewhere they don't have the 'regexp-opt' function ...
;;
;;(defvar xref-keywords-regexp
;;  (eval-when-compile
;;	(concat "\\b"
;;			(regexp-opt
;;			 '("auto" "extern" "register" "static" "typedef" "struct"
;;			   "union" "enum" "signed" "unsigned" "short" "long"
;;			   "int" "char" "float" "double" "void" "volatile" "const"
;;			   "break" "continue" "do" "else" "for" "if" "return"
;;			   "switch" "while" "goto" "case" "default" "sizeof"
;;			   ) t )
;;			"\\b")
;;	))

(defvar xref-font-lock-compl-keywords
    (cons (cons xref-keywords-regexp '(xref-keyword-face))
	  '(("^\\([a-zA-Z0-9_ ]*\\):" xref-list-pilot-face))
	  )
    "Default expressions to highlight in Xref completion list modes.")

;;;;;;;;;;;;;;;;;;;;;;;;;  highlighting ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-font-lock-list-keywords (sym)
    (cons (cons xref-keywords-regexp '(xref-keyword-face))
	(cons (cons (concat "\\b" sym "\\b") '(xref-list-symbol-face))
	  '(("^.\\([^:]*:[0-9]*\\):" xref-list-pilot-face))
	  ))
)


(defun completion-symbol-highlight (begpos endpos)
  (let ((line) (sym) (spos) (mbeg) (mend))
	(setq line (buffer-substring begpos endpos))
	(setq spos (string-match "[^a-zA-Z0-9_]" line 0))
	(setq sym (substring line 0 spos))
	(setq mbeg (string-match (concat "\\b[a-zA-Z0-9_\\.]*\\b" sym "\\b") line spos))
	(if mbeg
		(progn
		  (setq mend (match-end 0))
		  (put-text-property (+ mbeg begpos) 
							 (+ mend begpos) 
							 'face 'xref-list-symbol-face)
		  ))
))

(defun hightlight-keywords (begpos endpos keywords-regexp keyword-face)
  (let ((buff) (bpos) (loop-flag) (mbeg) (mend)
		)
	(setq buff (buffer-substring begpos endpos))
	(setq bpos 0)
	(setq loop-flag t)
	(while loop-flag
	  (setq mbeg (string-match keywords-regexp buff bpos))
	  (if (eq mbeg nil)
		  (setq loop-flag nil)
		(setq mend (match-end 0))
		(put-text-property (+ mbeg begpos) (+ mend begpos) 'face keyword-face)
		(setq bpos mend)
	  ))
))

(defun xref-fontify-region (begpos endpos kw-font-list)
  (let ((kwl) (kw))
	(setq kwl kw-font-list)
	(while kwl
	  (setq kw (car kwl))
	  (hightlight-keywords begpos endpos (car kw) (car (cdr kw)))
	  (setq kwl (cdr kwl))
	  )
))

(defun line-hightlight (begpos endpos multiple-lines-flag hbeg-offset) 
  (let ((bp) (loop-flag) (cpos) (prev-line) (dp-pos)
		)
;;(insert (format "[ %d-%d ]" begpos endpos))
	(setq cpos (char-search begpos -1 newline-char))
	(if (< cpos 0) (setq cpos 0))
	(if xref-coloring
		(xref-fontify-region (+ cpos 1) endpos xref-fontification)
	  )
	(if (or xref-mouse-highlight xref-coloring)
		(progn
		  (while (< (+ cpos 1) endpos)
			(setq bp cpos)
			(setq loop-flag t)
			(while (and loop-flag 
						multiple-lines-flag
						(eq (char-after (+ bp 1)) ?\ )
						(eq (char-after (+ bp 2)) ?\ )
						)
			  (setq prev-line (char-search (- bp 1) -1 newline-char))
			  (if (< prev-line 1)
				  (setq loop-flag nil)
				(setq bp prev-line)
				)
			  )
			(setq cpos (char-search (+ cpos 1) 1 newline-char))
			(if xref-mouse-highlight
				(put-text-property (+ bp hbeg-offset) cpos 'mouse-face 'highlight)
			  )
			(if xref-coloring
				(completion-symbol-highlight (+ bp 1) cpos)
			  )
			)))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun file-directory-name (fname)
  (let ((spos) (res))
	(setq res nil)
	(if (eq (length fname) 0)
		(setq res nil)
	  (setq spos (- (length fname) 1))
	  (while (and (> spos 0) 
				  (not (eq (elt fname spos) ?/))
				  (not (eq (elt fname spos) ?\\)))
		(setq spos (- spos 1))
		)
	  (if (> spos 0)
		  (setq res (substring fname 0 spos))
		(setq res nil)
		)
	  (if (and (eq xref-platform 'windows)
			   (eq spos 3)
			   (eq (elt fname (- spos 1)) ?\:))
		  (setq res nil)
		))
	res
))

(defun xref-pass-to-fun-lpar ()
  (search-forward-regexp xref-pass-identifier-regexp)
  (backward-char)
  (search-forward-regexp "[^
	 ]")
)

(defun identifier-on-point ()
  (let ((cpoin (point)) (cid))
	(if (or (eq (char-after cpoin) ?#) (eq (char-after cpoin) ?$))
		(forward-char)
	  )
	(search-forward-regexp xref-pass-identifier-regexp)
	(backward-char)
	(push-mark)
	(setq cid (buffer-substring cpoin (point)))
	(goto-char cpoin)
	cid
))

(defun char-search (pos step char-code)
  (let ((res pos))
	(while (and (>= res 1) (< res (point-max)) 
				(not (eq (char-after res) char-code)))
	  (setq res (+ res step))
	  )
	res
))

(defun delete-pending-ident ()
  (let ((p (point)))
	(search-forward-regexp xref-pass-identifier-regexp)
	(if (not (eq (point) (+ (buffer-size) 1)))
		(backward-char)
	  )
	(if (eq p (point))
		()
	  (delete-char (- p (point)) t)
	  (message "** Pending identifier killed (C-y can restore it) **")
	  )
))

(defun xrefsrv-filter (process output)
  (let ((lenn) (tmp-pos))
	(setq lenn (length output))
	(if (eq lenn 0) ()
	  (if (not filter-initialized)
		  (progn 
			(setq filter-initialized t)
			(setq xref-pilot-char (string-to-char (substring output 0 1)))
			(setq output (substring output 1))
			(cond
			 (
			  (eq xref-pilot-char ?\;)
				  (get-buffer-create " *xref-list*")
				  (set-buffer " *xref-list*")
				  (setq buffer-read-only nil)
				  (erase-buffer)
				  (setq xref-list-buffer-content nil)
			 )(
			  (eq xref-pilot-char ?\')
				  (get-buffer-create " *xref-execute*")
				  (set-buffer " *xref-execute*")
				  (setq buffer-read-only nil)
				  (erase-buffer)
			 )(
			  (eq xref-pilot-char ?\<)
				  (get-buffer-create "*xref-class-tree*")
				  (set-buffer "*xref-class-tree*")
				  (setq buffer-read-only nil)
				  (erase-buffer)
			 )(
			  (eq xref-pilot-char ?\>)
				  (get-buffer-create " *xref-list*")
				  (set-buffer " *xref-list*")
				  (setq buffer-read-only nil)
				  (erase-buffer)
				  (setq xref-list-buffer-content nil)
				  (insert-string 
				   (concat
					"Select references to process:   "
					"t)oggle (' '), a)ll, n)one, g)o (<CR>);\n"
					"Filter: p)lus('+'), m)inus('-')\n"
				   ))
			 )(
				(or (eq xref-pilot-char ?\.) (eq xref-pilot-char ?\,) 
					(eq xref-pilot-char ?\-) (eq xref-pilot-char ?\*)
					(eq xref-pilot-char ?\_) (eq xref-pilot-char ?\=)
					)
					nil
			 )(t
			   (get-buffer-create " *xrefsrv*")
			   (set-buffer " *xrefsrv*")
			   (setq buffer-read-only nil)
			   (erase-buffer)
			   )
			  )
			)
		)
   (if (eq xref-last-message-filtered nil)
	   (progn
		 (setq xref-last-message-filtered 
			   (string-match xref-sync-last-message output))
		 (if (not (eq xref-last-message-filtered nil))
			 (setq output (substring output 0 xref-last-message-filtered))
		   ))
	 )
   (setq lenn (length output))
   (if (eq lenn 0) ()
   (cond
	(
	(eq xref-pilot-char ?\,) 
	(set-buffer xrefsrv-buffer)
	(insert-string output)
	(delete-completion-window)
	(message " ")
	(delete-pending-ident)
   )(
	(eq xref-pilot-char ?\.) 
	(set-buffer xrefsrv-buffer)
	(insert-string output)
	(delete-completion-window)
	(delete-pending-ident)
	(message "** Multiple completions **")
   )(
	(eq xref-pilot-char ?\:)
	(set-buffer " *xrefsrv*")
	(display-buffer " *xrefsrv*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(insert output)
	(line-hightlight tmp-pos (point) t 1)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\!)
	(set-buffer " *xrefsrv*")
	(display-buffer " *xrefsrv*")
	(goto-char (point-max))
	(insert output)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\%)
	(set-buffer " *xrefsrv*")
	(display-buffer " *xrefsrv*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(insert output)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\~)
	(set-buffer " *xrefsrv*")
	(goto-char (point-max))
	(insert output)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\>)
	(set-buffer " *xref-list*")
	(display-buffer " *xref-list*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(setq buffer-read-only nil)
	(insert output)
	(line-hightlight tmp-pos (point) nil 2)
	(goto-char (point-min))
	(setq buffer-read-only t)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\<)
	(set-buffer "*xref-class-tree*")
	(goto-char (point-max))
	(setq buffer-read-only nil)
	(insert output)
	(goto-char (point-min))
	(setq buffer-read-only t)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\; )
	(set-buffer " *xref-list*")
	(display-buffer " *xref-list*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(setq buffer-read-only nil)
	(insert output)
	(line-hightlight tmp-pos (point) nil 1)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\' )
	(set-buffer " *xref-execute*")
	(goto-char (point-max))
	(insert output)
	(delete-completion-window)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\#)
	(set-buffer " *xrefsrv*")
	(insert-string output)
	(delete-completion-window)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\*) 
	(set-buffer xrefsrv-buffer)
	(setq xref-last-message output)
	(message "%s" output)
   )(
	(eq xref-pilot-char ?\-) 
	(set-buffer xrefsrv-buffer)
	(delete-completion-window)
	(message "** No completion possible **")
   )(
	(eq xref-pilot-char ?\_) 
	(set-buffer xrefsrv-buffer)
	(delete-completion-window)
	(message "** No reference **")
   )(
	(eq xref-pilot-char ?\=) 
	(set-buffer xrefsrv-buffer)
	(delete-completion-window)
	(message "** Empty stack, no pushed references **")
   )(
	t
	(set-buffer " *xrefsrv*")
	(display-buffer " *xrefsrv*")
	(message "**************** error in input format *****************")
	(insert-string output)
	(set-buffer xrefsrv-buffer)
   )
  )))
))

(defun start-xrefsrv-proc ()
  "sklsdkzxkzxk"
  (let ((oldpct process-connection-type))
	(setq process-connection-type nil)
	(setq xrefsrv-process (apply 'start-process 
                                 (concat xref-home-directory "xref") 
                                 nil 
                                 (concat xref-home-directory "xref")
								 "-task_regime_server"
                                 nil))
	(set-process-filter xrefsrv-process 'xrefsrv-filter)
	(setq process-connection-type oldpct)
  )
  (process-kill-without-query xrefsrv-process)
)

(defun xref-kill-xref-process (coredump)
"Kill currently running  xref process. This will kill  an xref process
creating/updating tag file. If there  is no such process this function
kills  the  xref server  process  answering  completion, browsing  and
refactoring  requests.  It  can  be used  when  xref task  goes in  an
inconsistent state.   Xref process is then  restarted automatically at
next  invocation  of any  of  its  functions.   The optional  argument
COREDUMP if present causes that the  core of the process is dumped for
debugging purposes.
"
;;  (interactive (list (if current-prefix-arg 'full)))
;; if previous do not work, use the next line rather
  (interactive "P")
  (if (and (not (eq xrefxref-process nil))
		   (eq (process-status xrefxref-process) 'run))
	  (progn
        (delete-process xrefxref-process)
        (setq xrefxref-process nil)
		(message "Extern Xref process killed.")
        )
	(setq xrefxref-process nil)
	(if (not (eq xrefsrv-process nil))
		(progn
		  (if current-prefix-arg
			  (progn
				(shell-command (format "kill -3 %d && echo Core dumped into this buffer directory." (process-id xrefsrv-process)))
				)
			(delete-process xrefsrv-process)
			(setq xrefsrv-process nil)
			(message "Emacs Xref server process killed.")
		  ))
	  (message "** No process to kill. **")
	  ))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-delete-window (buffername)
  (let ((displayed (get-buffer-window buffername nil)))
    (if displayed
        (progn
          (delete-window displayed)
		  (kill-buffer buffername)
        )
    )
  )
)

(defun delete-completion-window ()
  "sklsdkzxkzxk"
  (interactive "")
  (xref-delete-window  " *xrefsrv*")
  (xref-delete-window  "*background-1*")
  (xref-delete-window  " *xref-execute*")
)


(defun xref-delete-windows ()
  "Delete Emacs/XEmacs windows created by Xref."
  (interactive "")
  (delete-completion-window)
  (xref-delete-window  " *xref-list*")
  (xref-delete-window  "*xref-class-tree*")
  (xref-delete-window  "*xref-run*")
  (delete-other-windows)
  (message "Done.")
)

(defun write-xref-tmp-buff (tmpFile)
  (let ((inhib-o t))
	(if (boundp 'inhibit-eol-conversion)
		(progn
		  (setq inhib-o inhibit-eol-conversion)
		  (setq inhibit-eol-conversion t)
		  ))
	(write-region 1 (+ (buffer-size) 1) tmpFile nil 0)
	(if (boundp 'inhibit-eol-conversion)
		(setq inhibit-eol-conversion inhib-o)
	  )
	)
)

(defun xref-execute-command-is-complete ()
  (let ((res) (marker))
	(save-excursion
	  (set-buffer " *xref-execute*")
	  (beginning-of-buffer)
	  (end-of-line)
	  (setq marker (buffer-substring 1 (point)))
	  (setq res (search-forward marker nil t))
	  )
	res
))

(defun xref-execute-elisp-buffer ()
  (let ((mpos))
	(set-buffer " *xref-execute*")
	(setq mpos (xref-execute-command-is-complete))
	(goto-char mpos)
	(beginning-of-line)
	(kill-line)
	(beginning-of-buffer)
	(kill-line)
	(eval-buffer " *xref-execute*" nil)
))

(defun call-completion (sync save-flag poin mark callString arg1 arg2 arg3)
  (let ((originFile (buffer-file-name))
	    (originBuff (buffer-name))
	    (originPos (point))
		(original-size (+ (buffer-size) 1))
		(tmpFile xref-communication-file)
		(multibyte (and xref-allow-multibyte
						(boundp 'enable-multibyte-characters) 
						(not (eq nil enable-multibyte-characters))))
	   )
  (if (eq xrefsrv-process nil)
	  (start-xrefsrv-proc)
	(if (eq (process-status xrefsrv-process) 'run)
		()
	  (delete-process xrefsrv-process)
	  (start-xrefsrv-proc)
	)
  )
  (setq xrefsrv-buffer originBuff)
  (if (eq save-flag 0) 
	  ()
	(if multibyte
		(progn
		  (get-buffer-create " XrefTmpSaveBuffer")
		  (set-buffer " XrefTmpSaveBuffer")
		  (erase-buffer)
		  (insert-buffer-substring originBuff 1 original-size)
		  (goto-char poin)
		  (insert callString)
		  (write-xref-tmp-buff tmpFile)
		  (kill-buffer nil)
		  )
	  (write-xref-tmp-buff tmpFile)
	  )
	)
  (get-buffer-create " CompletionBuffer")
  (set-buffer " CompletionBuffer")
  (erase-buffer)
  (if (eq save-flag 0) 
	  ()
	(insert-string tmpFile)
	(newline)
	)
  (if xref-debug-mode
	  (progn
		(insert-string "-debug")
		(newline)	  
		))
  (if (eq originFile nil) 
	  ()
	(insert-string "-file")
	(newline)	  
	(insert-char ?" 1) (insert-string originFile) (insert-char ?" 1) 
	(newline)
	(if (not multibyte)
		(progn
		  (insert-string (format "-olcursor=%d" (- poin 1)))
		  (newline)
		  (insert-string (format "-olmark=%d" (- mark 1)))
		  (newline)
		  ))
;;	(insert-string "-olnodialog")
;;	(newline)
	(if xref-next-push-manual
		(progn
		  (insert-string "-olmanualresolve")
		  (newline)
		  (setq xref-next-push-manual nil)
		  ))
	)
  (insert-string "-user ")
  (insert-string xref-user-identification)
  (newline)
  (insert-string "-emacs ")
  (newline)
  (insert-string (format "-olinelen=%d" (window-width)))
  (newline)
  (if xref-current-project
	  (progn
		(insert-string (format "-p \"%s\"" xref-current-project))
		(newline)
		))
  (insert-string arg1)
  (newline)	  
  (insert-string arg2)
  (newline)	  
  (insert-string arg3)
  (newline)	  
  (if (eq sync 'sync)
	  (progn
		(insert-string (format "-last_message=%s" xref-sync-last-message))
		(newline)	  
		(setq xref-last-message-filtered nil)
		)
	(setq xref-last-message-filtered 0)
	)
  (newline)	  
  (insert-string "end-of-options")
  (newline)	  
  (newline)	  
  (setq filter-initialized nil)
  (setq xref-last-referencing-action 'xref-process-called)
  (process-send-region xrefsrv-process 1 (point))
  (erase-buffer)
  (kill-buffer nil)
  (get-buffer-create originBuff)
  (set-buffer originBuff)
  (message "")
  (accept-process-output xrefsrv-process)
  ;; is the output an e-lisp command ?
  (if (eq sync 'sync)
	  (progn
		;; synchronous dialog, wait until the end of answer
		(while (eq xref-last-message-filtered nil)
		  (accept-process-output xrefsrv-process)
		  ))
	;; assync, but even here you have to wait under some circumstance
	(if (eq xref-pilot-char ?\!)
		;; unexpected error, wait before delete the communication file
		(while (accept-process-output xrefsrv-process 1))
	  )
	(if (eq xref-pilot-char ?\')
		(progn
		  (while (not (xref-execute-command-is-complete))
			(accept-process-output xrefsrv-process)
			)
		  ))
	)
;;  the following can cause problem, because 'xrefsrv' in debug
;;  can output some message before processing input file, so it passes
;;  the 'accept-process-output' and deletes input, before opening it.
;;  FROM now, deleting is in the invocation procedures, however the above
;;  problem persist.
  (if (eq save-flag 0) 
	  ()
	(xref-delete-communication-file)
	)
  (if (eq xref-pilot-char ?\')
	  (xref-execute-elisp-buffer)
	)
))

(defun xref-delete-communication-file ()
  (if (not xref-debug-mode)
	  (delete-file xref-communication-file)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interactive-xref-completion-select (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno 1) (replace-flag))
	(if (eq (point) (point-max))
		(setq replace-flag nil)
	  (setq replace-flag t)
	  (setq lineno (count-lines 1 (+ (point) 1)))
	  )
	(other-window -1)
	(call-completion 'assync 0 0 0 "" (format "-olcomplselect%d" lineno)
					 "-olcxrefsuffix=0:#" "")
	(xref-moveref t)
	(if replace-flag
		(progn
		  (insert-string (substring xref-move-refid 1))
		  (delete-pending-ident)
		  ))
	(xref-delete-windows)
	(message " ")
	t
))

(defun interactive-xref-completion-mouse-select (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-select)
  t
)

(defun interactive-xref-completion-escape (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(call-completion 'assync 0 0 0 "" (format "-olcomplselect%d" lineno)
					 "-olcxrefsuffix=0:#" "")
	(xref-moveref t)
	(xref-delete-windows)
	(message " ")
	t
))

(defun interactive-xref-completion-mouse-escape (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-escape)
  t
)

(defun interactive-xref-completion-goto (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(call-completion 'assync 0 0 0 "" (format "-olcxcgoto%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(xref-moveref t)
;;	(put-text-property (point) (mark t) 'face 'highlight)
	(other-window 1)
  )
  t
)

(defun interactive-xref-completion-mouse-goto (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-goto)
  (other-window -1)
  t
)


(defun get-url-to-browse ()
  (let ((url "index.html"))
  (save-excursion
	(set-buffer " *xrefsrv*")
	(goto-char (point-min))
	(next-line 1)
	(setq url (buffer-substring 1 (- (point) 1)))
	)
;; following is because Netscape (as well as IE?) can't parse
;; urls with ')' and ',' while using remote control (IE also #?)
  (if (string-match "[,)#]" url)
	  (setq url (substring url 0 (string-match "#" url)))
	)
  url
))

(defun xref-browse-tmp-file-url (url)
  (browse-url url)
;;  (delete-file url)
)

(defun interactive-xref-completion-javadoc (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((url "index.html") (lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(call-completion 'assync 0 0 0 "" (format "-olcxcbrowse%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\~)
		  (xref-browse-tmp-file-url (get-url-to-browse))
	  )
	(other-window 1)
  )
  t
)

(defun interactive-xref-completion-mouse-javadoc (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-javadoc)
  (other-window -1)
  t
)


(defvar xref-completion-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-completion-escape)
    (define-key map "\e\e\e" 'interactive-xref-completion-escape)
    (define-key map "\C-m" 'interactive-xref-completion-select)
    (define-key map " " 'interactive-xref-completion-goto)
    (define-key map "j" 'interactive-xref-completion-javadoc)
    (define-key map "J" 'interactive-xref-completion-javadoc)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-completion-mouse-goto)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3up 'interactive-xref-completion-mouse-select)
		  (define-key map 'button3 'undefined)
		  (define-key map '(shift button2up) 'interactive-xref-completion-mouse-javadoc)
		  (define-key map '(shift button2) 'undefined)
		  (define-key map '(shift button3up) 'interactive-xref-completion-mouse-javadoc)
		  (define-key map '(shift button3) 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-completion-mouse-goto)
	  (define-key map [mouse-3] 'interactive-xref-completion-mouse-select)
	  (define-key map [S-mouse-2] 'interactive-xref-completion-mouse-javadoc)
	  (define-key map [S-mouse-3] 'interactive-xref-completion-mouse-javadoc)
	)
    map)
  "Keymap for `xref-completion-list-mode'."
)

(defvar xref-tag-search-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-completion-escape)
    (define-key map "\e\e\e" 'interactive-xref-completion-escape)
    (define-key map "\C-m" 'interactive-xref-completion-select)
    (define-key map " " 'interactive-xref-completion-goto)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-completion-mouse-goto)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3up 'interactive-xref-completion-mouse-select)
		  (define-key map 'button3 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-completion-mouse-goto)
	  (define-key map [mouse-3] 'interactive-xref-completion-mouse-select)
	)
    map)
  "Keymap for `xref-completion-list-mode'."
)

(defun xref-fontify-list (compl symbol)
  (let ((list-keywords))
;;	(font-lock-mode 1)
;;	(make-local-variable 'font-lock-defaults)
;;	(if compl
;;		(setq font-lock-defaults '(xref-font-lock-compl-keywords nil nil))
;;	  (setq list-keywords (xref-font-lock-list-keywords symbol))
;;	  (setq font-lock-defaults '(list-keywords nil nil))
;;	  )
;;	(font-lock-fontify-buffer)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-completion (poin)
"Complete  identifier.   This  function  parses the  currently  edited
buffer until the POIN position.  Then if the identifier on POIN have a
unique  completion it completes  it. If  there is  a list  of possible
completions they are displayed in a separate window.
"
  (interactive "d")
  (let ((copt ""))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (if xref-java-completion-linkage-check
		  (setq copt "-olchecklinkage")
		)
	  (if (not xref-java-completion-access-check)
		  (setq copt (concat copt " -olnocheckaccess"))
		)
	  (setq xref-fontification xref-font-lock-compl-keywords)
	  (call-completion 'assync 1 poin 0 "C@$./@mVeDitznACc" "-olcxcomplet" 
					   copt "-olcxrefsuffix=0:#")
	  (if (eq xref-pilot-char ?\;)
		  (progn
			(set-buffer " *xref-list*")
			(display-buffer " *xref-list*")
			(use-local-map xref-completion-mode-map)
;;			(message "button2 (' ') - inspect;  button3 (<CR>) - select;")
			(message "button2 (' ') - inspect;  button3 (<RET>) - select;  S-button2 ('j') - JavaDoc")
			)
		)
	  (if (eq xref-pilot-char ?\:)
		  (progn
			(set-buffer " *xrefsrv*")
			(display-buffer " *xrefsrv*")
			(use-local-map xref-completion-mode-map)
;;			(message "button2 (' ') - inspect;  button3 (<CR>) - select;")
			(message "button2 (' ') - inspect;  button3 (<RET>) - select;  S-button2 ('j') - JavaDoc")
			)
		)
;;	  (xref-delete-communication-file)
	  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-search-in-completions (poin)
"Search  string(s)  in  completions.   This function  first  asks  for
string(s) to search, then it  parses the currently edited buffer until
the  POIN  position.   Then  it  inspects  all  contextually  adequate
completions and displays only those containing the given string(s).
"
  (interactive "d")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(setq sstr (xref-get-search-string))
	(setq xref-fontification xref-font-lock-compl-keywords)
	(call-completion 'assync 1 poin 0 "C@$./@mVeDitznACc" 
					 "-olcxsearch" "-olcxrefsuffix=0:#"
					 (concat "\"-olcxcplsearch=" sstr "\"")
					 )
	(if (eq xref-pilot-char ?\;)
		(progn
		  (set-buffer " *xref-list*")
		  (display-buffer " *xref-list*")
		  (use-local-map xref-completion-mode-map)
		  (message "button2 (' ') - inspect;  button3 (<CR>) - select")
		  ;;	(xref-delete-communication-file)
		  ))
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun create-escaped-string (sstr)
  (let ((ind) (maxi) (ress) (cc))
	(setq ress "")
	(setq ind 0)
	(setq maxi (length sstr))
	(while (< ind maxi)
	  (setq cc (substring sstr ind (+ ind 1)))
	  (if (string-match "[A-Za-z_]" cc)
		  (setq ress (concat ress cc))
		(setq ress (concat ress "\\" cc))
		)
	  (setq ind (+ ind 1))
	  )
	ress
))

(defun create-xref-tag-search-fontif (sstr)
  (let ((res) (ress) (loop) (ind) (ind0) (prefix))
	(setq prefix "\\(")
	(setq loop t)
	(setq ind 0)
	(while loop
	  (setq ind0 (string-match "[^ \n\t]" sstr ind)) 
	  (if (not ind0)
		  (setq loop nil)
		(setq ind (string-match "[ \t\n]" sstr ind0))
		(if (not ind)
			(setq loop nil)
		  (setq ress (concat ress prefix (create-escaped-string (substring sstr ind0 ind))))
		  (setq prefix "\\|")
		  ))
	)
	(if ind0
		(setq ress (concat ress prefix (create-escaped-string (substring sstr ind0))))
	  )
	(setq ress (concat ress "\\)"))
	(setq res 
		  (cons (cons ress '(xref-list-symbol-face))
		  (cons (cons ":.*$" '(xref-list-default-face))
				nil))
		  )
	res
))

(defun xref-get-identifier-on-point ()
  (let ((sym))
	(setq poin (point))
	(search-backward-regexp xref-pass-identifier-regexp)
	(forward-char 1)
	(setq sym (identifier-on-point))
	(goto-char poin)
	sym
))

(defun xref-get-search-string() 
  (let ((sstr) (sym) (poin))
	(setq sym (xref-get-identifier-on-point))
	(setq sstr (read-from-minibuffer (format "String(s) to search [%s]:" sym)))
;;(error "'%s'" sstr)
	(if (equal sstr "")
		(setq sstr sym)
	  )
	sstr
))

(defun xref-current-column()
  (let ((poin) (res))
	(setq poin (point))
	(beginning-of-line)
	(setq res (- poin (point)))
	(goto-char poin)
	res
))


(defun xref-search-in-tag-file (poin)
"Search string(s) in the tag  file.  This function asks for strings to
search,  then  it inspects  the  Xref  tag  file looking  for  symbols
containing  given string(s).   For each  such symbol  it  displays the
symbol name together with its definition place.
"
  (interactive "d")
  (let ((sstr) (line) (col))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq sstr (xref-get-search-string))
	  (setq xref-fontification (create-xref-tag-search-fontif sstr))
	  (search-backward-regexp xref-pass-identifier-regexp)
	  (forward-char 1)
	  (setq line (number-to-string (count-lines (point-min) (+ (point) 1))))
	  (setq col (number-to-string (xref-current-column)))
	  (call-completion 'assync 1 poin 0 ""
					   (concat "-olcxlccursor=" line ":" col)
					   (concat "\"-olcxtagsearch=" sstr "\"")
					   "-olcxrefsuffix=0:#")
	  (set-buffer " *xref-list*")
	  (display-buffer " *xref-list*")
	  (use-local-map xref-tag-search-mode-map)
	  (message "button2 (' ') - inspect;  button3 (<CR>) - select")
;;	  (xref-delete-communication-file)
	  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun extract-process-address-arguments (headBeginPos headFinPos)
  (let ((addrlist nil) (sym) (poin)
		)
	(goto-char headBeginPos)
	(while (search-forward "@@" headFinPos t)
	  (delete-backward-char 2)
	  (setq addrlist (cons (point) addrlist))
	  )
	(while addrlist
	  (goto-char (car addrlist))
	  (setq sym (identifier-on-point))
	  (setq poin (point))
	  (map-on-all-references poin nil ''none-pre-fun ''renaming 
							 ''none-post-fun (concat "(*" sym ")") 
							 '(xref-popCxrefs poin)
							 )
	  (setq addrlist (cdr addrlist))
	  )
	)
)

(defun add-macro-line-continuations (reg-beg-pos reg-end-pos)
  (let ( (unmoved) (loopfl))
	(goto-char reg-end-pos)
	(setq loopfl t)
	(while loopfl
	  (end-of-line 1)
	  (insert "\\")
	  (setq unmoved (forward-line -1))
	  (setq loopfl (and (eq unmoved 0) (>= (point) reg-beg-pos)))
	  )
))

(defun extract-region (poin mark nname style)
  (let ((fcall) (fhead) (ftail) (fbody) (line) (cbuf) (pp) (addarg)
		(headBeginPos) (headFinPos) (reg-beg-pos) (reg-end-pos)
		(fun-call-pos))
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(if (eq style 'macro)
		(setq addarg "-olexmacro")
	  (if (eq style 'address)
		  (setq addarg "-olexaddress")
		(setq addarg "")
		)
	  )
	(call-completion 'sync 1 poin mark "" (concat "-olcxextract=" nname) addarg "")
	(if (not (eq xref-pilot-char ?%))
		()
	  (if (yes-or-no-p "Can I process the extraction ? ")
		  (progn 
			(setq cbuf (buffer-name))
			(setq fbody (buffer-substring poin mark))
			(set-buffer " *xrefsrv*")
			(goto-char (point-min))
			(search-forward "!")
			(search-forward "!")
			(setq pp (point))
			(search-forward "!")
			(setq fcall (buffer-substring pp (- (point) 1)))
			(search-forward "!")
			(setq pp (point))
			(search-forward "!")
			(setq fhead (buffer-substring pp (- (point) 1)))
			(search-forward "!")
			(setq pp (point))
			(search-forward "!")
			(setq ftail (buffer-substring pp (- (point) 1)))
			(setq pp (point))
			(search-forward "!")
			(setq line (string-to-int (buffer-substring pp (- (point) 1))))
			(set-buffer cbuf)
			(kill-region mark poin)
			(setq fun-call-pos (point))
			(insert-string fcall)
			(newline)
			(goto-line line)
			(setq headBeginPos (point))
			(insert-string fhead)
			(setq headFinPos (point))
			(if (not (eq style 'macro))
				(newline)
			  )
			(setq reg-beg-pos (point))
			(yank)
			(setq reg-end-pos (point))
			(insert-string ftail)
			(xref-delete-windows)
			(if (eq style 'macro)
				(add-macro-line-continuations reg-beg-pos reg-end-pos)
;;			  (extract-process-address-arguments headBeginPos headFinPos)
			  (goto-char headFinPos)
			  (if (yes-or-no-p "Extracted. Re-indent now ? ")
				  (progn
					(c-indent-defun)
;;	not work, as the position moves with the insertion of the new fun
;;					(goto-char fun-call-pos)
;;					(c-indent-defun)
					)
				)
			  )
			(message "Done.")
			)
		(xref-delete-windows)
		(message "** No extraction **")
		))
;;	(xref-delete-communication-file)
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-function (poin mark nname)
"This  function  parses  the  currently edited  buffer,  makes  static
analysis of  the C function  containing region between POIN  and MARK,
determines classes  of all local  variables, then it moves  the region
between the MARK and POIN before the function definition and generates
a head and the  tail of a new function around the  text of the region,
so it creates a new  function definition.  The generated function will
be  named NNAME.   The extraction  is  achieved by  inserting a  NNAME
function invocation at the original position of the region.
"
  (interactive "d
m
sthe name of the new function: ")
  (xref-multifile-undo-set-buffer-switch-point)
  (extract-region poin mark nname 'function)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-function-force-address (poin mark nname)
"This  function  parses  the  currently edited  buffer,  makes  static
analysis of  the C function  containing region between POIN  and MARK,
determines classes  of all local  variables, then it moves  the region
between the MARK and POIN before the function definition and generates
a head and the  tail of a new function around the  text of the region,
so it  creates a new function  definition. This function  will have no
return value, the  out parameters will be passed  as adress arguments.
The  generated  function  will  be  named NNAME.   The  extraction  is
achieved  by inserting  a NNAME  function invocation  at  the original
position of the region.
"
  (interactive "d
m
sthe name of the new function: ")
  (xref-multifile-undo-set-buffer-switch-point)
  (extract-region poin mark nname 'address)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-macro (poin mark nname)
"This  function  parses  the  currently edited  buffer,  makes  static
analysis of  the C function  containing region between POIN  and MARK,
determines classes  of all local  variables, then it moves  the region
between the MARK and POIN before the function definition and generates
a head and the  tail of a new macro around the  text of the region, so
it creates a new macro  definition.  The generated macro will be named
NNAME.  The extraction is achieved  by inserting a NNAME invocation at
the original position of the region.
"
  (interactive "d
m
sthe name of the new macro: ")
  (xref-multifile-undo-set-buffer-switch-point)
  (extract-region poin mark nname 'macro)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pass-fun-argument (num)
  (let ((deep 0)
		(nn num)
		(loop (> num 0))
		(ccc)
		)
	(while loop
	  (search-forward-regexp "[,\(\)\{\}/\"\']")
	  (setq ccc (char-after (- (point) 1)))
	  (cond
	   ((eq ccc ?\')
		(backward-char)
		(search-forward-regexp "[^\\]'"))
	   ((eq ccc ?\")
		(backward-char)
		(search-forward-regexp "[^\\]\""))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*))
		(search-forward "*/"))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/))
		(search-forward "
"))
	   ((and (eq ccc ?,) (eq deep 0))
		(setq nn (- nn 1))
		(setq loop (not (eq nn 0))))
	   ((and (eq ccc ?\)) (eq deep 0))
		(setq loop nil)
		(setq nn (- nn 1))
		(if (not (eq nn 0))
			(error "Out of arguments.")
		  ))
	   ((or (eq ccc ?\( ) (eq ccc ?\{))
		(setq deep (+ deep 1)))
	   ((or (eq ccc ?\) ) (eq ccc ?\}))
		(setq deep (- deep 1)))
	   )
	  )
	nn
))

(defun mark-fun-argument (num)
  (let ((rpoin) (res nil))
	(if (< num 1)
		(error "Argument number out of range")
	  (pass-fun-argument (- num 1))
	  (if (eq (char-after (- (point) 1)) ?\) )
		  (error "Argument number out of range")
		(setq rpoin (point))
		(pass-fun-argument 1)
		(backward-char)
		(push-mark)
		(goto-char rpoin)
		(setq res t)
		(message "Mark set at the end of the argument (C-w will delete it).")
		))
	res
))


(defun xref-moveref (check-id-and-go-arg)
  (let ((ffile) (fpoin) (fline) (fcoll) (fargn) (currentid) (refid) (rpos)
		(result nil))
	(if (not (eq xref-pilot-char ?#))
		()
	  (set-buffer " *xrefsrv*")
	  (if (< (buffer-size) 2)
		  ()
		(goto-char (point-min))
		(search-forward "#*+*#") 
		(setq ffile (buffer-substring 1 (- (point) 5)))
		(setq fpoin (point))
		(search-forward " ") 
		(setq fline (string-to-int (buffer-substring fpoin (- (point) 1))))
		(setq fpoin (point))
		(search-forward " ") 
		(setq fcoll (string-to-int (buffer-substring fpoin (- (point) 1))))
		(search-forward ":") 
		(setq xref-move-usage (char-after (point)))
		(forward-char 1)
		(setq fpoin (point))
		(search-forward ":") 
		(setq fargn (string-to-int (buffer-substring fpoin (- (point) 1))))
		(setq fpoin (point))
		(search-forward " ;;") 
		(setq refid (buffer-substring fpoin (- (point) 3)))
		(setq xref-move-refid refid)
		;;(message ffile) 
		;;(read-char)
		(kill-buffer nil)
		(xref-find-file ffile)
		(goto-line fline)
		(beginning-of-line)
		;;(message (concat fline "  " fcoll ))
		(setq rpos (+ (point) fcoll))
		(goto-char rpos)
;;(message "move-ref to %s %d %d" ffile fline fcoll)
		(setq lastref-file ffile)
		(setq lastref-line fline)
		(setq lastref-coll fcoll)
		(setq currentid (identifier-on-point))
		(message "Mark set at the end of the symbol.")
;;(setq transient-mark-mode t)
;;(setq deactivate-mark nil)
;;(setq mark-active t)
		(if check-id-and-go-arg
			(progn
			  (if (and (not (equal currentid  refid))
					   (not (equal (substring refid 0 1) "#")))
				  (error (concat "Hmm. Not a '"
								 refid
								 "' symbol reference. Please, update tag file."))
				(setq result t)
				(if (> fargn 0)
					(progn
					  (xref-pass-to-fun-lpar)
					  (if (not (eq (char-after (- (point) 1)) ?\( ))
						  (progn
							(error "Hmm. Does not seem as a function symbol reference. The argument not found.")
							(goto-char rpos)
							(setq result nil)
							)
						(setq result (mark-fun-argument fargn))
					  ))))))))
	result
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun show-xref-filter-level ()
  (setq mode-name (format "Xref-List Filter:%d" xref-list-filter-level))
)

(defun show-xref-symbol-filter-level ()
  (setq mode-name (format "Xref-Menu Filter:%d" xref-sym-menu-filter-level))
)

(defun interactive-xref-list-goto (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((poin (point)) (lastref) (lineno (count-lines 1 (+ (point) 1))))
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if lastref
		(progn
		  (delete-backward-char 1)
		  (insert " ")
		  )
	  )
	(goto-line lineno)
	(beginning-of-line)
	(delete-char 1)
	(insert ">")
	(goto-char poin)
	(other-window -1)
	(call-completion 'assync 0 0 0 "" (format "-olcxgoto%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(xref-moveref t)
	(other-window 1)
  )
  (message xref-list-buffer-message)
  t
)

(defun next-reference-list-buffer ()
  (let ((cline) (nlines) (poin (point)) (lastref))
	(if (not (eq xref-list-buffer-content 'reference-list))
		(error "Not a reference list buffer")
	  )
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if (not lastref)
		(goto-char poin)
	  )
	(beginning-of-line)
	(setq cline (count-lines (point-min) (point)))
	(setq nlines (count-lines (point-min) (point-max)))
	(setq cline (+ cline 1))
	(if (< cline nlines)
		(progn
		  (next-line 1)
		  (interactive-xref-list-goto)
		  )
	  (message "* last reference *")
;;	  (goto-char (point-min))
	  )
))

(defun previous-reference-list-buffer ()
  (let ((cline) (nlines) (poin (point)) (lastref))
	(if (not (eq xref-list-buffer-content 'reference-list))
		(error "Not a reference list buffer")
	  )
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if (not lastref)
		(goto-char poin)
	  )
	(beginning-of-line)
	(setq cline (count-lines (point-min) (point)))
	(setq nlines (count-lines (point-min) (point-max)))
	(setq cline (- cline 1))
	(if (>= cline 0)
		(progn
		  (next-line -1)
		  (interactive-xref-list-goto)
		  )
	  (message "* first reference *")
;;	  (beginning-of-line nlines)
	  )
))

(defun interactive-xref-list-filter-plus (&optional argp)
  "Increase the view filter level"
  (interactive "P")
  (let ((fopt))
	(if (< xref-list-filter-level 3)
		(progn
		  (setq xref-list-filter-level (+ xref-list-filter-level 1))
		  (setq fopt (format "-olcxfilter=%d" xref-list-filter-level))
		  (other-window -1)
		  (call-completion 'assync 0 0 0 "" fopt "-olcxrefsuffix=0:" "")
		  (other-window 1)
		  (show-xref-filter-level)
		  ))
	(message xref-list-buffer-message)
  t
))

(defun interactive-xref-list-filter-minus (&optional argp)
  "Decrease the view filter level"
  (interactive "P")
  (let ((fopt))
	(if (> xref-list-filter-level 0)
		(progn
		  (setq xref-list-filter-level (- xref-list-filter-level 1))
		  (setq fopt (format "-olcxfilter=%d" xref-list-filter-level))
		  (other-window -1)
		  (call-completion 'assync 0 0 0 "" fopt "-olcxrefsuffix=0:" "")
		  (other-window 1)
		  (show-xref-filter-level)
		  ))
	(message xref-list-buffer-message)
	t
))

(defun interactive-xref-list-mouse-goto (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-list-goto)
  (other-window -1)
  t
)

(defun interactive-xref-pop (&optional argp)
  (interactive "P")
  (xref-popCxrefs (point))
  t
)

(defun interactive-xref-mouse-pop (event)
  (interactive "e")
  (mouse-set-point event)
  (xref-popCxrefs (point))
  t
)

(defvar xref-list-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-pop)
    (define-key map "\e\e\e" 'interactive-xref-pop)
    (define-key map "\C-m" 'xref-delete-windows)
    (define-key map " " 'interactive-xref-list-goto)
    (define-key map "+" 'interactive-xref-list-filter-plus)
    (define-key map "-" 'interactive-xref-list-filter-minus)
    (define-key map [kp-add] 'interactive-xref-list-filter-plus)
    (define-key map [kp-subtract] 'interactive-xref-list-filter-minus)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-list-mouse-goto)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3up 'interactive-xref-mouse-pop)
		  (define-key map 'button3 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-list-mouse-goto)
	  (define-key map [mouse-3] 'interactive-xref-mouse-pop)
	)
    map)
  "Keymap for `xref-list-mode'."
)

(defun set-xref-listrefs-dialog-maps ()
  (set-buffer " *xref-list*")
  (display-buffer " *xref-list*")
  (select-window (get-buffer-window " *xref-list*"))
  (show-xref-filter-level)
  (use-local-map xref-list-mode-map)
  (setq xref-list-buffer-content 'reference-list)
  (other-window -1) 
  (message xref-list-buffer-message)
)

(defun handle-pushrefs-action (action)
  (eval action)
;;  (if (eq action 'moveto) 
;;	  (xref-moveref t)
;;	(if (eq action 'listrefs)
;;		(if (eq xref-pilot-char ?\; )
;;			(set-xref-listrefs-dialog-maps)
;;		  )))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun interactive-xref-menu-selection-select (&optional argp)
  "Select the symbol corresponding to this line."
  (interactive "P")
  (let ((lineno) (str) (nstr) (poin))
	(setq poin (point))
	(setq lineno (count-lines 1 (+ (point) 1)))
	(if (> lineno xref-menu-selection-prefix-lines)
		(progn
		  (beginning-of-line)
		  (setq str (buffer-substring (point) (+ (point) 1)))
		  (if (or (string-equal str " ") (string-equal str "+"))
			  (progn
				(if (string-equal str " ")
					(setq nstr "+")
				  (setq nstr " ")
				  )
				(setq buffer-read-only nil)
				(delete-char 1)
				(insert-string nstr)
				(setq buffer-read-only t)
				(goto-char poin)
				(next-line 1)
				;; (other-window -1)
				(call-completion 'assync 0 0 0 "" (format "-olcxmenuselect%d" lineno) 
								 "-olcxrefsuffix=0:" "")
				;; (other-window 1)
				))
		  ))
  )
  t
)

(defun apply-fun-menu-selection-lines (fromline toline fun)
  (let ((count) (str))
	(goto-line fromline)
	(setq count fromline)
	(while (<= count toline)
	  (progn
		(beginning-of-line)
		(setq str (buffer-substring (point) (+ (point) 1)))
		(if (or (string-equal str " ") (string-equal str "+"))
			  (eval fun)
		  )
		(next-line 1)
		(setq count (+ count 1))
		))
))
  

(defun interactive-xref-menu-selection-all (&optional argp)
  "Select all the symbols."
  (interactive "P")
  (save-excursion
	(apply-fun-menu-selection-lines 
	 (+ xref-menu-selection-prefix-lines 1)
	 (count-lines 1 (point-max))
	 '(progn 
		(beginning-of-line)
		(setq buffer-read-only nil)
		(delete-char 1)
		(insert "+")
		(setq buffer-read-only t)
		))
	(call-completion 'assync 0 0 0 "" "-olcxmenuall" "-olcxrefsuffix=0:" "")
	)
  t
)

(defun interactive-xref-menu-selection-none (&optional argp)
  "Unselect all the symbols."
  (interactive "P")
  (save-excursion
	(apply-fun-menu-selection-lines 
	 (+ xref-menu-selection-prefix-lines 1)
	 (count-lines 1 (point-max))
	 '(progn 
		(beginning-of-line)
		(setq buffer-read-only nil)
		(delete-char 1)
		(insert " ")
		(setq buffer-read-only t)
		))
	(call-completion 'assync 0 0 0 "" "-olcxmenunone" "-olcxrefsuffix=0:" "")
	)
  t
)

(defun interactive-xref-menu-selection-mouse-select (event)
  (interactive "e")
  (save-excursion
	(mouse-set-point event)
	(interactive-xref-menu-selection-select)
	)
  t
)

(defun interactive-xref-menu-selection-go (&optional argp)
  (interactive "P")
  (xref-delete-windows)
  (call-completion 'assync 0 0 0 "" "-olcxmenugo" "-olcxrefsuffix=0:" "")
  (handle-pushrefs-action xref-pushrefs-action)
  t
)

(defun interactive-xref-menu-selection-mouse-go (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-menu-selection-go)
  t
)

(defun interactive-xref-menu-filter-plus (&optional argp)
  (interactive "P")
  (let ((fopt))
	(setq buffer-read-only nil)
	(erase-buffer)
	(if (< xref-sym-menu-filter-level 2)
		(setq xref-sym-menu-filter-level (+ xref-sym-menu-filter-level 1)))
	(setq fopt (format "-olcxmenufilter=%d" xref-sym-menu-filter-level))	
	(call-completion 'assync 0 0 0 "" fopt "-olcxrefsuffix=0:" "")
	(handle-ambiguous-browsing-dialog xref-pushrefs-action)
	t
))

(defun interactive-xref-menu-filter-minus (&optional argp)
  (interactive "P")
  (let ((fopt))
	(setq buffer-read-only nil)
	(erase-buffer)
	(if (> xref-sym-menu-filter-level 0)
		(setq xref-sym-menu-filter-level (- xref-sym-menu-filter-level 1)))
	(setq fopt (format "-olcxmenufilter=%d" xref-sym-menu-filter-level))	
	(call-completion 'assync 0 0 0 "" fopt "-olcxrefsuffix=0:" "")
	(handle-ambiguous-browsing-dialog xref-pushrefs-action)
	t
))

(defvar xref-menu-selection-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'xref-delete-windows)
    (define-key map "\C-m" 'interactive-xref-menu-selection-go)
    (define-key map "g" 'interactive-xref-menu-selection-go)
    (define-key map "G" 'interactive-xref-menu-selection-go)
    (define-key map " " 'interactive-xref-menu-selection-select)
    (define-key map "t" 'interactive-xref-menu-selection-select)
    (define-key map "T" 'interactive-xref-menu-selection-select)
    (define-key map "a" 'interactive-xref-menu-selection-all)
    (define-key map "A" 'interactive-xref-menu-selection-all)
    (define-key map "n" 'interactive-xref-menu-selection-none)
    (define-key map "N" 'interactive-xref-menu-selection-none)
    (define-key map "p" 'interactive-xref-menu-filter-plus)
    (define-key map "P" 'interactive-xref-menu-filter-plus)
    (define-key map "+" 'interactive-xref-menu-filter-plus)
    (define-key map [kp-add] 'interactive-xref-menu-filter-plus)
    (define-key map "m" 'interactive-xref-menu-filter-minus)
    (define-key map "M" 'interactive-xref-menu-filter-minus)
    (define-key map "-" 'interactive-xref-menu-filter-minus)
    (define-key map [kp-subtract] 'interactive-xref-menu-filter-minus)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 
			'interactive-xref-menu-selection-mouse-select)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3up 'interactive-xref-menu-selection-mouse-go)
		  (define-key map 'button3 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-menu-selection-mouse-select)
	  (define-key map [mouse-3] 'interactive-xref-menu-selection-mouse-go)
;;	  (define-key map [menu-bar] (make-sparse-keymap))
;;	  (define-key map [menu-bar xref-view]
;;		(cons "Xref-View" (make-sparse-keymap "Xref-View")))
;; 	  (define-key map [menu-bar xref-view name]
;;		'("Symbols of the same name" . xref-menu-view-name))
;; 	  (define-key map [menu-bar xref-view profil]
;;		'("Symbols with the same profil" . xref-menu-view-profil))
;; 	  (define-key map [menu-bar xref-view profil]
;;		'("Symbols virtualy adequate" . xref-menu-view-virtual))
	)
    map)
  "Keymap for xref-menu for symbol selection."
)

(defun handle-ambiguous-browsing-dialog (action)
  (set-buffer " *xref-list*")
  (goto-char (point-min))
;; this should be also accessible by the standard menu pop-up
  (display-buffer " *xref-list*")
  (select-window (get-buffer-window " *xref-list*"))
  (goto-line (+ xref-menu-selection-prefix-lines 3))
  (setq xref-pushrefs-action action)
  (use-local-map xref-menu-selection-mode-map)
  (show-xref-symbol-filter-level)
  (setq buffer-read-only t)
  (message "button2 (' ') - select/unselect;  button3 (<CR>) - go")
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun push-references (poin pusharg bargnum continuation)
  (let ((name)
		(modified) (line) (col) (tbpars "")
		)
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq modified (buffer-modified-p))
	  (search-backward-regexp xref-pass-identifier-regexp)
	  (if (not (eq (char-after (point)) ?#))
		  (forward-char 1)
		)
	  (setq name (identifier-on-point))
	  (if modified
		  (setq tbpars "-olcxmodified")
		(setq line (count-lines (point-min) (+ (point) 1)))
		(setq col (xref-current-column))
		(setq tbpars 
			  (format "-olcxlccursor=%d:%d -olcxunmodified -browsedsym=%s"
					  line col name))
		)
	  (goto-char poin)
	  (if (and (string-equal name "")
			   (not (eq (point) (point-min))))
		  (message "No symbol, position the cursor on a symbol name first.")

		(xref-delete-windows)
		(setq xref-list-filter-level 0)
		(setq xref-sym-menu-filter-level 2)
		(call-completion 'assync 1 poin 0 "C@$./@mVeDitznACx" pusharg 
						 (format "-olcxrefsuffix=%d:%s" bargnum name)
						 tbpars)
		(if (eq xref-pilot-char ?\> )
			(handle-ambiguous-browsing-dialog continuation)
		  (handle-pushrefs-action continuation)
		  )
;;		(xref-delete-communication-file)
))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pushing-post-action ()
  (let ((url) (answer))
	(if (eq xref-pilot-char ?\~)
		(progn
		  ;; no definition; browse javadoc URL
		  (setq url (get-url-to-browse))
		  (if xref-ask-before-browse-javadoc
			  (progn
				(message 
				 "No definition reference found. Browse JavaDoc [y/n]?")
				(setq answer (get-single-touch-event))
				(message "")
				)
			(setq answer 'act)
			)
		  (if (eq answer 'act)
			  (xref-browse-tmp-file-url url)
			)
		  )
	  (xref-moveref t)
	  )
))


(defun xref-pushCxrefs (poin)
"This function causes that the  currently edited buffer is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected,  its  references are  pushed  on  the  reference stack,  the
references are then completed from the Xref tag (cross reference) file
and  finally the  point is  moved to  the definition  of  the recorded
symbol.
"
  (interactive "d")
  (push-references poin "-olcxpush" 0 '(pushing-post-action))
)


(defun xref-push-references (poin)
"This function causes that the  currently edited buffer is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected, its  references are pushed  on the reference stack,  and the
references  are then  completed from  the Xref  tag  (cross reference)
file.
"
  (interactive "d")
  (push-references poin "-olcxpushonly" 0 nil)
)

(defun pushCxrefsForRename (poin)
  "push references and go to the definition"
  (interactive "d")
  (push-references poin "-olcxrename" 0 '(xref-moveref t))
)

;;;;;;;;;;;;;;;;;;;;;;;;;; Class Tree ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interactive-xref-class-tree-frame (&optional argp)
  (interactive "P")
  (if (not window-system)
	  (if (not (yes-or-no-p 
				"Really create new frame on charecter based terminal ?"))
		  (error "Exiting, no frame created")
		))
  (other-window -1)
  (delete-other-windows)
  (make-frame)
  (other-frame -1)
  (switch-to-buffer "*xref-class-tree*")
)

(defun xref-scroll-left (&optional argp)
  (interactive "P")
  (scroll-left 1)
)

(defun xref-scroll-right (&optional argp)
  (interactive "P")
  (scroll-right 1)
)


(defvar xref-class-tree-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'xref-delete-windows)
    (define-key map [left] 'xref-scroll-right)
    (define-key map [right] 'xref-scroll-left)
    (define-key map "f" 'interactive-xref-class-tree-frame)
    (define-key map "F" 'interactive-xref-class-tree-frame)
    map)
  "Keymap for xref class tree."
)


(defun class-tree-dialog ()
  (let ((ctw))
	(if (eq xref-pilot-char ?\<)
		(progn
		  (set-buffer "*xref-class-tree*")
		  (goto-char (point-min))
		  (setq truncate-lines t)
		  (setq ctw (get-buffer-window "*xref-class-tree*" t))
		  (if (not ctw)
			  (progn
				(display-buffer "*xref-class-tree*")
				(setq ctw (get-buffer-window "*xref-class-tree*" t))
				)
			)
		  (select-window ctw)
		  (use-local-map xref-class-tree-mode-map)
		  (setq buffer-read-only t)
		  (message "'f' - new frame")
		))
))


(defun xref-class-tree-show (poin)
"This function  is menaingfull only  for Java programmers.   It causes
that the currently edited buffer  is parsed, the references of symbols
are recorded, the symbol on the  POIN position is resolved. If it is a
class, then a class hierarchy consisting from its sub and superclasses
is showed. By decreasing filter level you can get the full class tree.
If the symbol under  point is not a class but a  class member then the
class tree  is showed  for the class  containing this member.   If the
symbol has  no class, then the  full hierarchy starting  by the Object
class is showed.
"
  (interactive "d")
  (push-references poin "-olcxclasstree" 0 '(class-tree-dialog))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; List ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-listCxrefs (poin)
"This function causes that the  currently edited buffer is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected,  its  references are  pushed  on  the  reference stack,  the
references are then completed from the Xref tag (cross reference) file
and all references are showed in a separate window.
"
  (interactive "d")
  (let ((iden) (opos))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (save-excursion
		(get-buffer-create " *xref-list*")
		(set-buffer " *xref-list*")
		(erase-buffer)
		)
	  (setq opos (point))
	  (search-backward-regexp xref-pass-identifier-regexp)
	  (forward-char 1)
	  (setq iden (identifier-on-point))
	  (goto-char opos)
	  (setq xref-fontification (xref-font-lock-list-keywords iden))
	  (setq xref-list-filter-level 0)
;;	  (push-references poin "-olcxlist -olmanualresolve" 0 
	  (push-references poin "-olcxlist" 0 '(listrefs-continuation))
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; pop refs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-show-stack-top-symbol (poin)
"Display the name of the symbol on the top of reference stack. This is
usually a fully qualified name.
"
  (interactive "d")
  (call-completion 'assync 0 (point) 0 "" "-olcxtops" "" "")
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; pop refs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-popCxrefs (poin)
"Pop one symbol references from the reference stack and move point 
at the position from where the references were pushed.
"
  (interactive "d")
  (xref-delete-windows)
  (call-completion 'assync 0 poin 0 "" "-olcxpop" "" "")
  (if (eq xref-pilot-char ?#)
	  (progn
		(xref-moveref nil)
		(xref-show-stack-top-symbol (point))
		))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; re-push refs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-re-push (poin)
"Pop one symbol references from the reference stack and move point 
at the position from where the references were pushed.
"
  (interactive "d")
  (xref-delete-windows)
  (call-completion 'assync 0 poin 0 "" "-olcxrepush" "" "")
  (if (eq xref-pilot-char ?#)
	  (progn
		(xref-moveref nil)
		(xref-show-stack-top-symbol (point))
		))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;; LOCAL MOTION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-this-symbol-local-motion (poin continuation)
  (let ((iden) (opos))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq opos (point))
	  (search-backward-regexp xref-pass-identifier-regexp)
	  (forward-char 1)
	  (setq iden (identifier-on-point))
	  (goto-char opos)
	  (push-references poin "-olcxpushonly" 0 continuation)
)))

(defun xref-poponly ()
  "Pop references from the top of the stack."
  (interactive "")
  (xref-delete-windows)
  (call-completion 'assync 0 (point) 0 "" "-olcxpoponly" "" "")
)


(defun xref-this-symbol-previous-ref (poin)
  "Go to the previous reference of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (xref-previousref) 
				  (call-completion 'assync 0 (point) 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)


(defun xref-this-symbol-next-ref (poin)
  "Go to the next reference of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (xref-nextref) 
				  (call-completion 'assync 0 (point) 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)

(defun xref-this-symbol-definition-ref (poin)
  "Go to the definition of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (call-completion 'assync 0 0 0 "" "-olcxgotodef" 
								   "-olcxrefsuffix=0:" "")
				  (xref-moveref t)
				  (call-completion 'assync 0 0 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defun xref-nextref ()
  "Go to the next reference on the top of the reference stack."
  (interactive "")
  (let ((list-win) (cwin) (compile-win))
	(if (string-equal (buffer-name) " *xref-list*")
		(next-reference-list-buffer)
	  (setq list-win (get-buffer-window " *xref-list*"))
	  (setq compile-win (get-buffer-window "*compilation*"))
	  (if (and compile-win (eq xref-last-referencing-action 'compilation))
		  (next-error 1)
		(if (and list-win (eq xref-list-buffer-content 'reference-list))
			(progn
			  (setq cwin (selected-window))
			  (select-window list-win)
			  (next-reference-list-buffer)
			  (select-window cwin)
			  )
		  (call-completion 'assync 0 0 0 "" "-olcxplus" "-olcxrefsuffix=0:" "")
		  (xref-moveref t)
)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-previousref ()
  "Go to the previous reference on the top of the reference stack."
  (interactive "")
  (let ((list-win) (cwin))
	(if (string-equal (buffer-name) " *xref-list*")
		(previous-reference-list-buffer)
	  (setq list-win (get-buffer-window " *xref-list*"))
	  (setq compile-win (get-buffer-window "*compilation*"))
	  (if (and compile-win (eq xref-last-referencing-action 'compilation))
		  (next-error -1)
		(if (and list-win (eq xref-list-buffer-content 'reference-list))
			(progn
			  (setq cwin (selected-window))
			  (select-window list-win)
			  (previous-reference-list-buffer)
			  (select-window cwin)
			  )
		  (call-completion 'assync 0 0 0 "" "-olcxminus" "-olcxrefsuffix=0:" "")
		  (xref-moveref t)
)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-simple-output-filter (process output)
  (let ((buffer (current-buffer))
		(wind (selected-window))
		(scroll scroll-step)
		(owin)
		)
	(setq scroll-step 1)
;;	(set-buffer "*background-1*")
	(setq owin (get-buffer-window "*background-1*"))
	(if owin
		(progn
		  (select-window owin)
		  (goto-char (point-max))
		  (insert output)
		  (sit-for 0)
		  (setq scroll-step scroll)
		  ;;	(set-buffer buffer)
		  (select-window wind)
		  )
	  (message "%s" output)
	  )
))

(defun xref-call-non-interactive-process (updateoption sync wind)
  (let ((buffer (current-buffer)) 
		(bdir default-directory)
		(ubuf) (args))
    (setq ubuf (generate-new-buffer "*background-1*"))
	(display-buffer "*background-1*")
	(setq default-directory bdir)
	(setq args nil)
 	(if xref-current-project
		(setq args (append (list "-p" xref-current-project) args))
	  )
	(setq args (append (list "-last_message=Done." updateoption) args))
	(setq xrefxref-process (apply 'start-process "xref-assync" nil
								  (concat xref-home-directory "xref") args))
	(set-process-filter xrefxref-process 'xref-simple-output-filter)
	(if (eq sync 'synchronized)
		(progn
		  (message "Please wait.")
		  (while (eq (process-status xrefxref-process) 'run)
			(progn
			  (accept-process-output xrefxref-process)
			  ;;(sit-for 0)
			  ))
		  )
	  (message "Please wait, then You can delete the new window by hitting F7.")
	  )
))


(defun xref-update-refs ()
"Update Xref  cross reference file.  This function  execute the 'xref'
command with the  -update option. If the current  project is set (with
the  'xref-set-current-project' function) then  also the  '-p project'
option is set. The effect of the invocation is that all files modified
since last  update (but  not taking into  account 'fast  updates' (see
also 'xref-fast-update-refs'  function) are  processed as well  as all
files  (recursively) including  those modified  files.  In  result the
cross reference  file is safely  updated.  The behavior of  the 'xref'
command is controlled by options read from the '.xrefrc' file.
"
  (interactive "")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(xref-call-non-interactive-process "-update" 'assync 'newwin)
))


(defun xref-fast-update-refs ()
"Update Xref  cross reference file.  This function  execute the 'xref'
command with  the -fastupdate  option. If the  current project  is set
(with  the  'xref-set-current-project'  function)  then also  the  '-p
project' option  is set.  The effect  of the invocation  is that files
modified since  last (fast) updating  of the cross reference  file are
processed and  the cross references are  updated.  No include-included
relation is  taken into account  when determining which files  will be
processed.  Under  a very specific  conditions, the 'fast  update' can
lead  to  a lost  or  misunderstanding of  few  references,  so it  is
recommended  to   make  full   update  before  any   important  source
manipulating  operation.   The  behavior  of  the  'xref'  command  is
controlled by options read from the '.xrefrc' file.
"
  (interactive "")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(xref-call-non-interactive-process "-fastupdate" 'assync 'newwin)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-create-refs ()
"Create Xref  cross reference file.  This function  execute the 'xref'
command with the  -create option. If the current  project is set (with
the  'xref-set-current-project' function) then  also the  '-p project'
option is set. The effect of  the invocation is that a cross reference
file is created,  the behavior of the 'xref'  command is controlled by
options read from the '.xrefrc' file.
"
  (interactive "")
  (xref-call-non-interactive-process "-create" 'assync 'newwin)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun  xref-about ()
  "Show Xref version and current project setting informations."
  (interactive "")
  (xref-call-non-interactive-process "-about" 'synchronized 'newwin)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;       MAPPINGS        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ref-equal (file1 file2 line1 line2 coll1 coll2) 
  (and (string-equal file1 file2) (equal line1 line2) (equal coll1 coll2)))

(defun get-single-yes-no-event () 
  (let ((key) (res nil))
	(while (eq res nil)
	  (setq key (apply xref-get-event nil))
	  (setq key (vector key))
;;	  (setq key (read-key-sequence-vector "")
	  (setq res (lookup-key xref-query-replace-map key))
	  (if (eq res nil)
		  (message "Invalid key, please answer [y/n].")
		)
	  )

	res
))

(defun get-single-touch-event () 
  (let ((key) (res nil))
	(setq key (apply xref-get-event nil))
	(setq key (vector key))
;;	  (setq key (read-key-sequence-vector "")
	(setq res (lookup-key xref-query-replace-map key))
	res
))

(defun check-and-count-references ()
  (let ((f0) (l0) (c0) (ff) (ll) (cc) (refcheck) (loop t) (num 0))
	(message "Checking references")
	(xref-previousref)
	(setq f0 lastref-file)
	(setq l0 lastref-line)
	(setq c0 lastref-coll)
	(while loop
	  (setq refcheck (xref-previousref))
	  (if (not refcheck)
		  (progn
			(setq loop nil)
			(setq num 0)
			)
		(message "Checking references")
		(setq ff lastref-file)
		(setq ll lastref-line)
		(setq cc lastref-coll)
		(setq num (+ num 1))
		(setq loop (not (ref-equal f0 ff l0 ll c0 cc)))
	  ))
  num
))

(defun none-pre-fun (arg1 arg2)
  nil
)

(defun none-post-fun (arg1 arg2)
  nil
)

(defun standard-post-fun (interactive rrr)
  (if (and interactive (eq rrr 'act))
	  (progn
		(message 
		 "The occurence is processed. Hit any key to pass to the next reference.")
		(apply xref-get-event nil)
;;		(get-single-touch-event)
		)
	)
)

;;;;;;;;;;;;;;;;;;;; MAPPING FUNCTION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; - pre-fun is usually positioning point and mark on important positions
;; - fun provides the action
;; - post fun make a post action before move to next reference
;; !!!!! the mapped 'fun' returns its error message or nil for O.K. !!!!
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun map-on-all-references2 (poin interact pre-fun fun post-fun arg1 arg2)
  (let ((num) (count) (app) (mess "") (pass-ref) (rrr) (problem))
	(setq count (check-and-count-references))
	(setq num count)
	(if (eq num 0)
		(progn
;;		  (xref-popCxrefs poin)
		  (message "Something is wrong, maybe not updated tag file?")
		)
	  (xref-nextref)
	  (while (> num 0)
		(xref-previousref)
		(setq mess "Done.")
		(setq app (apply pre-fun arg1 arg2 nil))
		(if (not (eq app nil))
			(setq mess app)
		  )
		(if (and interact (eq app nil))
			(progn
			  (message " Proceed this occurence [y/n]? ")
			  (setq rrr (get-single-yes-no-event))
			  )
		  (setq rrr 'act)
		  )
		(if (and (eq app nil) (eq rrr 'act))
			(progn
			  (setq app (apply fun arg1 arg2 nil))
			  (undo-boundary)
			  (if (not (eq app nil))
				  (setq mess app)
				)
			  )
		  )
		(setq problem (not (eq app nil)))
		(while (and interact problem)
		  (message (concat mess " Type 'c' to confirm."))
		  (setq rrr (get-single-touch-event))
		  (setq problem (not (eq rrr 'confirm)))
		  )  
		(apply post-fun interact rrr nil)
		(setq num (- num 1))
	  )
;;	  (xref-popCxrefs poin)
	  (message "Done.")
	)
	count
))

(defun map-on-all-references (poin interact pre-fun fun post-fun arg1 arg2 
								   continuation)
  (push-references poin "-olcxrename" 0 
    (list 'progn
	   (list 'map-on-all-references2 poin interact pre-fun fun post-fun arg1 arg2)
	   (list 'eval continuation)
	   ))
)

(defun exchange-text (p1b p1e p2b p2e) 
  (let ((s1 (buffer-substring p1b p1e))
		(s2 (buffer-substring p2b p2e)))
	(if (< p1b p2b)
		(progn
		  (goto-char p2b)
		  (delete-char (- p2e p2b))
		  (insert s1)
		  (goto-char p1b)
		  (delete-char (- p1e p1b))
		  (insert s2)
		  )
	  (goto-char p1b)
	  (delete-char (- p1e p1b))
	  (insert s2)
	  (goto-char p2b)
	  (delete-char (- p2e p2b))
	  (insert s1)
	  )
	)
)

(defun exchange-arguments (arg1 arg2)
  "exchange of two arguments of a function"
  (let ((pos1beg)(pos1end)(pos2beg)(pos2end)(rrr)
		(pos (point)) (fspos) (result nil) (rflag1) (rflag2))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq fspos (point))
	  (setq rflag1 (mark-fun-argument arg1))
	  (setq pos1beg (point))
	  (setq pos1end (mark t))
	  (goto-char fspos)
	  (setq rflag2 (mark-fun-argument arg2))
	  (setq pos2beg (point))
	  (setq pos2end (mark t))
	  (if (and rflag1 rflag2)
		  (exchange-text pos1beg pos1end pos2beg pos2end)
		(setq result "Argument out of range.")
	  )
  )
  (goto-char pos)
  result
))


(defun xref-arg-exchange (poin)
"Exchange two arguments of a  function symbol at POIN of the currently
edited buffer.  The  function interactively asks for the  order of the
arguments you wish to exchange, then it inspects all references of the
symbol taken  from the tag file  checking them to really  point to the
symbol. If any  of the references is out of date,  the action will not
occur  and  the function  finishes  with  an  error message.   If  all
references are O.K.  the function passes throw the references one more
time exchanging strings of the selected arguments (under the condition
that the text around looks like a function definition/invocation).
"
  (interactive "d")
  (let ((name) (arg1) (arg2) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-multifile-undo-set-buffer-switch-point)
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to exchange [ 1 - arity('%s') ] : " name name) 
		 "1"
		 ))
	  (setq arg1 (string-to-int argns))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Exchange with argument #[ 1 - arity('%s') ] : " name) 
		 "2"
		 ))
	  (setq arg2 (string-to-int argns))
	  (map-on-all-references poin t ''none-pre-fun ''exchange-arguments 
						 ''standard-post-fun arg1 arg2 
						 '(xref-popCxrefs poin)
						 )
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun shift-arguments (arg1 arg2)
  "shift an argument of a function"
  (let ((pos1beg)(pos1end)(pos2)(rrr)(abpos)(s1)
		(pos (point)) (result nil) (rflag1) (rflag2))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq abpos (point))
	  (setq rflag1 (mark-fun-argument arg1))
	  (setq pos1beg (point))
	  (setq pos1end (mark t))
	  (goto-char abpos)
	  (setq rflag2 (mark-fun-argument arg2))
	  (setq pos2 (point))
	  (if (not (and rflag1 rflag2))
		  (setq result "Argument out of range.")
		(setq s1 (buffer-substring pos1beg pos1end))
		(goto-char pos1end)
		(delete-char (- pos1beg pos1end))
		(if (eq (char-after (- (point) 1)) ?\( )
			(if (eq (char-after (point)) ?\) )
				()
			  (delete-char 1)
			  )
		  (delete-char -1)
		  )
		(goto-char abpos)
		(if (eq arg2 1)
			(if (eq (char-after (point)) ?\) )
				(insert-string s1)
			  (insert-string (concat s1 ","))
			  )
		  (setq rflag2 (mark-fun-argument (- arg2 1)))
		  (goto-char (mark t))
		  (if (eq (char-after (point)) ?\) )
				(insert-string (concat "," s1))
			(forward-char 1)
			(insert-string (concat s1 ","))
			)
		  )
		))
	(goto-char pos)
	result
))


(defun xref-arg-shift (poin)
"Shift  an argument  of a  function symbol  at POIN  of  the currently
edited buffer.  The  function interactively asks for the  order of the
argument you wish  to shift and for the order where  you wish to shift
it.  Then it inspects all references  of the symbol taken from the tag
file  checking them  to really  point  to the  symbol. If  any of  the
references is out of date, the  action will not occur and the function
finishes  with an  error message.   If  all references  are O.K.   the
function passes throw  the references one more time  moving the string
of the selected argument to the new position (under the condition that
the text around looks like a function definition/invocation).
"
  (interactive "d")
  (let ((name) (arg1) (arg2) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-multifile-undo-set-buffer-switch-point)
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to shift [ 1 - arity('%s') ] : " name name) 
		 "1"
		 ))
	  (setq arg1 (string-to-int argns))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Shift to argument # [ 1 - arity('%s') ] : " name) 
		 "2"
		 ))
	  (setq arg2 (string-to-int argns))
	  (map-on-all-references poin t ''none-pre-fun ''shift-arguments 
						 ''standard-post-fun arg1 arg2
						 '(xref-popCxrefs poin)
						 )
)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pre-position-to-argument-delete (arg1 arg2)
  "position point and mark to the argument arg1 of a function"
  (let ((result nil) (rflag1))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq rflag1 (mark-fun-argument arg1))
	  (if (not rflag1)
		  (setq result "Argument out of range.")
		))
	result
))

(defun delete-argument (arg1 arg2)
  "delete the argument arg1 of a function"
  (let ((result nil))
	(delete-region (point) (mark t))
	(if (eq (char-after (- (point) 1)) ?\( )
		(if (eq (char-after (point)) ?\) )
			()
		  (delete-char 1)
		  )
	  (delete-char -1)
	  )
	result
))


(defun xref-arg-delete (poin)
"Delete  an argument of  a function  symbol at  POIN of  the currently
edited buffer.  The  function interactively asks for the  order of the
argument you  wish to delete, then  it inspects all  references of the
symbol taken  from the tag file  checking them to really  point to the
symbol. If any  of the references is out of date,  the action will not
occur  and  the function  finishes  with  an  error message.   If  all
references are O.K.  the function passes throw the references one more
time deleting the string of the selected argument (under the condition
that the text around looks like a function definition/invocation).
"
  (interactive "d")
  (let ((name) (arg) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-multifile-undo-set-buffer-switch-point)
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to delete [ 1 - arity('%s') ] : " name name) 
		 "1"
		 ))
	  (setq arg (string-to-int argns))
	  (map-on-all-references poin t ''pre-position-to-argument-delete
						 ''delete-argument ''standard-post-fun arg nil
						 '(xref-popCxrefs poin)
						 )
)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pre-position-to-argument-insert (arg1 arg2)
  "position point and mark to the argument arg1 of a function"
  (let ((result nil) (nn))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (if (eq arg1 0)
		  ()
		(setq nn (pass-fun-argument (- arg1 1)))
		(if (not (eq nn 0))
			(progn
			  (beep) 
			  (setq result "Argument out of range.")
			  )
		  )
		)
	  )
	result
))

(defun reference-insert-text (argtext)
  (if (or (eq xref-move-usage ?*)
		  (eq xref-move-usage ?+)
		  (eq xref-move-usage ?!))
	  (car argtext)
	(car (cdr argtext))
))

(defun insert-argument (arg1 argtext)
  "insert the argument arg1 of a function "
  (let ((pos))
	(if (eq (char-after (- (point) 1)) ?\) )
		(progn
		  (backward-char 1)
		  (insert-char ?, 1)
		  (insert (reference-insert-text argtext))
		  )
	  (insert (reference-insert-text argtext))
	  (setq pos (point))
	  (search-forward-regexp "[^ ]")
	  (if (eq (char-after (- (point) 1)) ?\) )
		  (goto-char pos)
		(goto-char pos)
		(insert-char ?, 1)
		)
	  )
	)
  nil
)


(defun xref-arg-insert (poin)
"Add an  argument to the function  symbol at the POIN  position of the
currently edited buffer.  This function  asks you for the order of the
argument you wish  to insert and for the string you  wish to insert as
default argument  value, then it  identifies the symbol on  the point,
pushes all  its references on the  reference stack, checks  if all the
references point really  to this symbol, if the  references are out of
date, the  function provides an  error message. If all  references are
O.K.  It  inspects all references  one more time, inserting  the comma
and the default string at the argument position.
"
  (interactive "d")
  (let ((name) (arg) (argns) (text) (textdef) (textval))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-multifile-undo-set-buffer-switch-point)
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to insert [ 1 - arity('%s') ] : " name name)
		 "1"
		 ))
	  (setq arg (string-to-int argns))
	  (setq textdef (read-from-minibuffer 
		 (format "String to insert as declaration of argument #%d: " arg) 
		 "int arg"
		 ))
	  (setq textval (read-from-minibuffer 
		 (format "String to insert as default value of argument #%d: " arg) 
		 "0"
		 ))
	  (setq text '(cons textdef (cons textval nil)))
	  (map-on-all-references poin t ''pre-position-to-argument-insert
						 ''insert-argument ''standard-post-fun arg text
						 '(progn
							(call-completion 'assync 0 poin 0 ""
											 (format "-olcxresetrefsuffix=%d:%s" arg name) 
											 "" "")
							(message " ** Done. References kept on stack for manual adjustements. **")
							)
						 )
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-arg-push (poin)
"Push references to  an argument of the function  symbol positioned at
the  POIN  position of  the  currently edited  buffer  and  go to  the
definition.  This function asks you  for the order of the argument you
wish to browse, then it identifies the symbol on the point, pushes all
its references on  the reference stack, moves on  the definition place
of the symbol and them moves to its n-th argument (under the condition
that the  the text here  looks like a function  definition/invocation) .
Then you  can inspect next/previous  reference of the symbol  with the
'xref-nextref' and 'xref-previousref'  functions. Each time the cursor
will be moved at the n-th argument position.
"
  (interactive "d")
  (let ((name) (argn) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-multifile-undo-set-buffer-switch-point)
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to  browse [ 1 - arity('%s') ] : " name) 
		 "1"
		 ))
	  (setq argn (string-to-int argns))
	  (push-references poin "-olcxpush" argn 
					   '(progn 
						  (xref-moveref t)
						  (message "Pushed. Use 'Previous ref' and 'Next refs' to inspect the references.")
						  )))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun renaming (new-name dummy)
  (let ((pos (point)))
	(kill-region (point) (mark t))
	(insert new-name)
	(goto-char pos)
	nil
))

(defun xref-definition-reference ()
"Find a definition reference on the top of the stack and go there."
  (interactive "")
  (call-completion 'assync 0 (point) 0 "" "-olcxgotodef" "" "")
  (xref-moveref t)
)

(defun xref-caller-reference ()
"Go to the position from where references on the top of the stack were pushed."
  (interactive "")
  (call-completion 'assync 0 (point) 0 "" "-olcxgotocaller" "" "")
  (xref-moveref t)
)

(defun xref-current-reference ()
"Go to the current reference of the stack."
  (interactive "")
  (call-completion 'assync 0 (point) 0 "" "-olcxgotocurrent" "" "")
  (xref-moveref t)
)

(defun listrefs-continuation ()
  (if (eq xref-pilot-char ?\; ) 
	  (set-xref-listrefs-dialog-maps)
	)
)


(defun xref-list-top-references ()
"List references on the top of the reference stack."
  (interactive "")
  (call-completion 'assync 0 (point) 0 "" "-olcxlisttop" "" "")
  (listrefs-continuation)
)

(defun xref-safety-check ()
  (let ((prompt "Done. Save modified files and proceed safety check? "))
	(if (yes-or-no-p prompt)
		(progn
		  (xref-save-some-buffers t)
		  (call-completion 'assync 0 (point) 0 "" "-olcxgotodef" "" "")
		  ;; goto definition without checking the symbol name
		  (xref-moveref nil)
		  (call-completion 'assync 0 (point) 0 "" "-olcxsafetycheck1" 
						   "-olcxrefsuffix=0:#" "")
		  (if (eq xref-pilot-char ?\*)
			  (progn
				(xref-call-non-interactive-process "-update" 'synchronized 'nowin)
				(message "Checking new references.")
				(push-references (point) "-olcxsafetycheck2" 0 
								 '(listrefs-continuation))
				)
			(listrefs-continuation)
			)
		  )
	  (message "** Changes unsaved, save buffers manually **")
	  )
))


(defun xref-rename (poin)
"Safe  renaming  of a  symbol  positionned  on  position POIN  of  the
currently edited  buffer. The function interactively asks  for the new
name, then it inspects all references of the symbol taken from the tag
file  checking them  to really  point  to the  symbol. If  any of  the
references is  out of date, the  renaming will not  occur and function
finishes  with an  error.  If  all references  are O.K.   the function
passes throw  the references  one more time  replacing the  old symbol
name with the new one.
"
  (interactive "d")
  (let ((tcount) (prompt) (new-name) (old-name))
	(xref-multifile-undo-set-buffer-switch-point)
	(setq old-name (xref-get-identifier-on-point))
	(setq new-name (read-from-minibuffer 
					(format "Rename '%s' to : " old-name) old-name))
	(setq tcount (map-on-all-references poin t ''none-pre-fun ''renaming 
										''none-post-fun new-name nil
										'(xref-safety-check)
))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-toggle-next-push-manual ()
"This function sets the  xref-next-push-manual flag to true. This will
cause that  the next reference pushing action  (push references, list,
rename,  arg.   manipulation, ...)  will  first  pop  a manual  symbol
resolution menu as if the symbol is ambiguous.  This can be usefull if
you wish to  pass around the automatic symbol  resolution and you wish
to get for example references of all the symbols of the given name
 (not only those of the symbol pointed by point)."
  (interactive "")
  (setq xref-next-push-manual t)
  (message "Next reference pushing action will used manual resolution.")
)


(defun xref-pushCxrefs-force-ctree (poin)
"This function  will force the  class tree resolution dialog  for next
push and will then call 'Push Refernces' function.
"
  (interactive "d")
  (xref-toggle-next-push-manual)
  (xref-pushCxrefs poin)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            Multifile UNDO
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar xref-multifile-undo-state nil)

(defun xref-this-buffer-undo-state ()
  (undo-boundary)
  (cdr buffer-undo-list)
)

(defun xref-editor-undo-state ()
  (let ((state) (buffs) (bst) (buf) (bnamestart))
	(save-excursion
	  (setq state nil)
	  (setq buffs (buffer-list))
;;	  (setq buffs (append buffs (cons (current-buffer) nil)))
	  (while buffs
		(setq buf (car buffs))
		(setq buffs (cdr buffs))
		(setq bnamestart (substring (buffer-name buf) 0 1))
		(if (or (equal bnamestart " ") (equal bnamestart "*"))
			;; do not process buffers starting by " " or "*"
			()
		  (set-buffer buf)
		  (setq bst (xref-this-buffer-undo-state))
		  (setq state (cons (cons buf bst) state))
		  )))
	state
))

(defun get-corresponding-undo (ulist buf)
  (let ((ul) (found))
	(setq ul ulist)
	(setq found nil)
	(while (and ul (not found))
	  (setq found (eq (car (car ul)) buf))
	  (if (not found)
		  (setq ul (cdr ul))
		)
	  )
	(if found
		(cdr (car ul))
	  nil
	  )
))	  

(defun finish-undoing (message)
  (if (yes-or-no-p (concat message "Save modified buffers? "))
	  (progn
		(xref-save-some-buffers t)
		(error "Changes saved.")
		)
	(error "Modifications not saved.")
))

(defun undo-changes-until (bundo)
  (let ((cc) (aa))
	(setq loop t)
	(while (and pending-undo-list (not (eq pending-undo-list bundo)))
	  (undo-more 1)
	  (message "Continue undoing [y/n]?")
	  (setq aa (get-single-yes-no-event))
	  (if (eq aa 'noact)
		  (finish-undoing "** Undoing breaked. ")
		)
	  (message "")
	  (undo-boundary)
	  )
	(if (not (eq pending-undo-list bundo))
		(error "No further undo information available (check 'undo-limit' and 'undo-strong-limit' variables)!") 
	  )
))

(defun xref-undo-single-buffer (buf ustate cont-ustate)
  (let ((bnamestart) (bufundo))
	(setq bnamestart (substring (buffer-name buf) 0 1))
	(if (or (equal bnamestart " ") (equal bnamestart "*"))
		;; do not process buffers starting by " " or "*"
		()
	  (switch-to-buffer buf)
	  (setq bufundo (get-corresponding-undo ustate buf))
	  (if cont-ustate
		  (setq pending-undo-list (get-corresponding-undo cont-ustate buf))
		(undo-start)
		(if bufundo (undo-more 1))
		)
	  (undo-changes-until bufundo)
	  )
))

(defun xref-undo-until-undo-state (ustate cont-ustate)
  (let ((buffs) (buf) (buffs2) (bufundo))
	(setq buffs (buffer-list))
	;; first process bufferes without undo-state
	(while buffs
	  (setq buf (car buffs))
	  (setq buffs (cdr buffs))
	  (setq bufundo (get-corresponding-undo ustate buf))
	  (if (not bufundo)
		  (xref-undo-single-buffer buf ustate cont-ustate)
		)
	  )
	;; that those from list
	(setq buffs2 ustate)
	(while buffs2
	  (setq buf (car (car buffs2)))
	  (setq buffs2 (cdr buffs2))
	  (setq bufundo (get-corresponding-undo ustate buf))
	  (if bufundo
		  (xref-undo-single-buffer buf ustate cont-ustate)
		)
	  )
))

(defun cut-long-refactoring-undo-list ()
  (let ((ul) (deep))
	(if (> (length xref-multifile-undo-state) xref-multifile-undo-deep)
		(progn
		  (setq ul xref-multifile-undo-state)
		  (setq deep xref-multifile-undo-deep)
		  (while (and ul (> deep 0))
			(setq deep (- deep 1))
			(setq ul (cdr ul))
			)
		  (if ul
			  (setcdr ul nil)
			))
	  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-undo-last-refactoring () 
"This function  undoes changes made in multiple  buffers.  Buffers are
switched at beginnings of  refactorings. The undo itself is considered
as a refactoring,  so further invocations can be  used to re-do undone
changes.   It uses  standard  Emacs  undo mecanism  which  may not  be
appropriate for  such a  massive undoing, in  case of  problems verify
your 'undo-strong-limit' and 'undo-limit' variable settings.
"
  (interactive "")
  (let ((undolist) (cloop) (cont-undo-state) (aa))
	(if (eq xref-multifile-undo-state nil)
		(error "** Refactorings undo stack is empty **")
	  )
	(save-excursion 
	  (setq cont-undo-state nil)
	  (setq cloop t)
	  (setq undolist xref-multifile-undo-state)
	  (while cloop
		(xref-multifile-undo-set-buffer-switch-point)
		(xref-undo-until-undo-state (car undolist) cont-undo-state)
		(setq cont-undo-state (car undolist))
		(setq undolist (cdr undolist))
		(if (eq undolist nil)
			(setq cloop nil)
		  (setq cloop (yes-or-no-p 
					   "A refactoring is undone, continue by previous one? "))
		  )
		)
	  (finish-undoing "Undone. ")
	  )
))


(defun xref-multifile-undo-set-buffer-switch-point ()
"This  function sets an  multifile undo  state-point, meaning  that it
memorize  the  state of  undo  buffers  from  all opened  buffers.  An
invocation of xref-undo-last-refactoring  will then undo changes under
those points in all those buffers.
"
  (interactive "")
  (setq xref-multifile-undo-state 
		(cons (xref-editor-undo-state) xref-multifile-undo-state))
  (cut-long-refactoring-undo-list)
;;(insert (format "multifile-undo-state : %S\n\n" xref-multifile-undo-state))
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;        IDE     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; this function can be used to run an xref user defined non-interactive
;; commands, the value of 'name' argument has to be set in .xrefrc file.

(defun xref-compile-function (name arg) 
  (require 'compile)
  (if (not (featurep 'compile))
	  (error "Your (X)Emacs do not contain the 'compile.el' package, install it first")
	)
  (call-completion 'assync 1 0 0 "" "-get" name
				   "-olcxrefsuffix=0:")
  (if (eq xref-pilot-char ?\*) 
	  (progn
		(compile (format xref-last-message arg))
;;		(message "F3/F4 will work for previous/next error")
		)
	)
  (setq xref-last-referencing-action 'compilation)
)

(defun xref-ide-compile-file () 
"Compile file. The  compile command must be specified  in your .xrefrc
file  via the '-set  compilefile <command>'  option. If  the <command>
string contains the %s character  sequence, it will be replaced by the
currently edited file name.  Read xref manual file for more info about
-set option.   This function actually only calls  the (X)Emacs compile
function with  the appropriate  command string. You  need to  have the
'compile' package installed in your editor.

Xref-Tech is not author of  compile.el package which is released under
GPL.
"
  (interactive "")
  (xref-compile-function "compilefile" (backslashify-name (buffer-file-name)))
)

(defun xref-ide-compile-dir () 
"Compile  directory. The  compile command  must be  specified  in your
.xrefrc  file  via the  '-set  compiledir  <command>'  option. If  the
<command>  string  contains the  %s  character  sequence,  it will  be
replaced by  the current  directory name.  Read  xref manual  file for
more info  about -set option.   This function actually only  calls the
(X)Emacs  compile function  with the  appropriate command  string. You
need to have the 'compile' package installed in your editor.

Xref-Tech is not author of  compile.el package which is released under
GPL.
"
  (interactive "")
  (xref-compile-function "compiledir" (backslashify-name (file-directory-name (buffer-file-name))))
)

(defun xref-ide-compile-project () 
"Compile  directory. The  compile command  must be  specified  in your
.xrefrc  file  via the  '-set  compiledir  <command>'  option. If  the
<command>  string  contains the  %s  character  sequence,  it will  be
replaced by the current project  name.  Read xref manual file for more
info about -set option. This function actually only calls the (X)Emacs
compile function with the appropriate command string. You need to have
the 'compile' package installed in your editor.

Xref-Tech is not author of  compile.el package which is released under
GPL.
"
  (interactive "")
  (call-completion 'assync 1 0 0 "" "-olcxgetprojectname" 
				   "-olcxrefsuffix=0:" "")
  (if (eq xref-pilot-char ?\*)
	  (xref-compile-function "compileproject" xref-last-message)
	(error "No project Selected")
	)
)

(defun xref-ide-previous-error ()
"Move to  the previous  error of the  last compilation.  This function
actually calls next-error with -1 argument. 
"
  (interactive "")
  (setq xref-last-referencing-action 'compilation)
  (next-error -1)
)

(defun xref-ide-next-error ()
"Move to  the previous  error of the  last compilation.  This function
actually calls next-error with -1 argument.
"
  (interactive "")
  (setq xref-last-referencing-action 'compilation)
  (next-error 1)
)

;; this function can be used to run an xref user defined interactive
;; commands, the value of 'name' argument has to be set in .xrefrc file.
(defun xref-run-function (name)
  (let ((bb) (rc) (args) (ww) (bdir) (cc))
	(setq bdir default-directory)
	(call-completion 'assync 1 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*)
		(progn
		  (require 'comint)
		  (if (not (featurep 'comint))
			  (error "Your (X)Emacs do not contain the 'comint.el' package, install it first")
			)
		  (call-completion 'assync 1 0 0 "" "-get" name
						   "-olcxrefsuffix=0:")
		  (if (eq xref-pilot-char ?\*) 
			  (progn
				(setq rc xref-last-message)
				(get-buffer-create " *xref-run-command*")
				(set-buffer " *xref-run-command*")
				(erase-buffer)
				(setq cc (format "cd %s\n%s\n" (backslashify-name bdir) rc))
				(insert-string cc)
				(write-region 1 (+ (buffer-size) 1) xref-run-batch-file)
				(kill-buffer nil)
				(if (eq xref-platform 'windows)
					(setq args (cons xref-run-batch-file nil))
				  (setq args (cons "sh" (cons xref-run-batch-file nil)))
				  )
				(get-buffer-create "*xref-run*")
				(setq ww (display-buffer "*xref-run*"))
				(select-window ww)
				(erase-buffer)
				(if (not (eq xref-platform 'windows))
					(insert-string cc)
				  )
				(setq default-directory bdir)
				(apply 'make-comint "xref-run" (car args) nil (cdr args))
				)
			)
		  )
	  (error "No project Selected")
	)
))

(defun xref-ide-run ()
"Run the command specified vith the '-set compiledir <command>' option
in your .xrefrc file.  Read xref  manual file for more info about -set
option.   This  function  actually  only  calls  the  (X)Emacs  comint
function with  the appropriate command  string.  You need to  have the
'comint' package installed in your editor.

Here follows comint key-bindings  taken from the comint documentation.
It  may be  somewhere  inaccurate, for  exact  documentation see  your
'comint.el'  source  file. 

;; Brief Command Documentation:
;;============================================================================
;; Comint Mode Commands: (common to all derived modes, like shell & cmulisp
;; mode)
;;
;; m-p     comint-previous-input           Cycle backwards in input history
;; m-n     comint-next-input               Cycle forwards
;; m-r     comint-previous-matching-input  Previous input matching a regexp
;; m-s     comint-next-matching-input      Next input that matches
;; m-c-l   comint-show-output              Show last batch of process output
;; return  comint-send-input
;; c-d     comint-delchar-or-maybe-eof     Delete char unless at end of buff
;; c-c c-a comint-bol-or-process-mark      First time, move point to bol;
;;                                         second time, move to process-mark.
;; c-c c-u comint-kill-input                ^u
;; c-c c-w backward-kill-word               ^w
;; c-c c-c comint-interrupt-subjob          ^c
;; c-c c-z comint-stop-subjob               ^z
;; c-c c-\ comint-quit-subjob               ^\
;; c-c c-o comint-kill-output              Delete last batch of process output
;; c-c c-r comint-show-output              Show last batch of process output
;; c-c c-l comint-dynamic-list-input-ring  List input history

Xref-Tech is not  author of comint.el package which  is released under
GPL.
"
  (interactive "")
  (xref-run-function "run")
)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;    Projects    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-get-project-list ()
  (let ((new-name) (loop) (mbeg) (mend) (pname "") (project-list))
	(set-buffer (get-buffer-create " xref-project-list"))
	(erase-buffer)
	(insert-file xref-options-file)
	(goto-char (point-min))
	(setq project-list nil)
	(setq loop t)
	(while loop
	  (setq loop (search-forward-regexp "\\[\\([^\]]*\\)\\]" 
										(buffer-size) 1))
	  (if loop
		  (progn
			(setq mbeg (match-beginning  1))
			(setq mend (match-end  1))
			(setq pname (buffer-substring mbeg mend))
			(setq project-list (cons pname project-list))
			)))
	(kill-buffer nil)
	project-list
))


(defun interactive-xref-project-select (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((bl) (el) (prj))
	(beginning-of-line)
	(setq bl (point))
	(end-of-line)
	(setq el (point))
	(setq prj (buffer-substring (+ bl 2) el))
	(if (string-equal prj xref-directory-dep-prj-name)
		(setq xref-current-project nil)
	  (setq xref-current-project prj)
	  )
	(xref-delete-window  " *xref-project-list*")
	(message "Current project: '%s'" prj)
  )
  t
)

(defun interactive-xref-m-project-select (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-project-select)
  (other-window -1)
  t
)

(defun interactive-project-escape (&optional argp)
  "Escape from the project selection window."
  (interactive "P")
  (xref-delete-window  " *xref-project-list*")
)

(defvar xref-project-list-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'interactive-project-escape)
    (define-key map "\C-m" 'interactive-xref-project-select)
    (define-key map " " 'interactive-xref-project-select)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-m-project-select)
		  (define-key map 'button2 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-m-project-select)
	)
    map)
  "Keymap for `xref-project-list-mode'."
)


(defun xref-project-set-active ()
"Set  active  project.   This  function  is meningfull  only  if  your
'.xrefrc'  file   contains  a   section  defining  options   for  your
project. After setting a project to be active all Xrefactory functions
will proceed according to options corresponding to this project name.
"
  (interactive "")
  (let ((new-name) (pname) (bl))
	(setq xref-project-list (xref-get-project-list))
	(get-buffer-create " *xref-project-list*")
	(set-buffer " *xref-project-list*")
	(erase-buffer)
	(insert-string "  ")
	(insert-string xref-directory-dep-prj-name)
	(put-text-property 3 (point) 'mouse-face 'highlight)
	(newline)
	(while xref-project-list
	  (goto-char (point-min))
	  (if (string-equal xref-current-project (car xref-project-list))
		  (insert-string "> ")
		(insert-string "  ")
		)
	  (insert-string (car xref-project-list))
	  (put-text-property 3 (point) 'mouse-face 'highlight)
	  (newline)
	  (setq xref-project-list (cdr xref-project-list))
	  )
	(display-buffer " *xref-project-list*")
	(use-local-map xref-project-list-mode-map)
	(message "Use button2, <space> or <CR> to select the project")
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun remove-pending-slash (pfiles)
  (let ((dlen))
	(setq dlen (- (length pfiles) 1))
	(if (or (eq (elt pfiles dlen) ?/) (eq (elt pfiles dlen) ?\\))
		(setq pfiles (substring pfiles 0 dlen))
	  )
	pfiles
))

(defun backslashify-name (fname) 
  (let ((spos) (len) (res))
	(setq res fname)
	(if (or (eq xref-platform 'unix) (eq (length fname) 0))
		(setq res fname)
	  (setq len (length fname))
	  (setq spos 0)
	  (while (< spos len) 
		(if (or (eq (elt fname spos) ?/) (eq (elt fname spos) ?\\))
			(store-substring fname spos xref-slash)
		  )
		(setq spos (+ spos 1))
		)
	  )
	res
))

(defun xref-append-new-project-section (pname planguage pcomments pfiles)
  (let ((comment) (dlen))
	(setq comment (or (equal pcomments "y") (equal pcomments "Y")))
	(end-of-buffer)
	(insert-string (concat "\n\n[" pname "]\n"))
	(if comment
		(insert-string "  // input files and directories\n")
	  )
	(insert-string (concat "  " pfiles))
	(newline)

	(setq pfiles (remove-pending-slash pfiles))
	(if comment
		(insert-string "  // directory where Tag files are stored\n")
	  )
	(insert-string (format "  -refs %s%cXrefs\n" pfiles xref-slash))

	(if comment
		(insert-string "  // number of Tag files\n")
	  )
	(insert-string "  -refnum=10\n")

	(if (or (equal planguage "c") (equal planguage "C")
			(equal planguage "b") (equal planguage "B"))
		(progn
		  (if (eq xref-platform 'windows)
			  (insert-string "  // -I c:\\includes -I c:\\home\\include   ")
			(insert-string "  // -I /usr/local/includes -I /home/include   ")
			)
		  (insert-string " // put include directories here\n")
		  )
	  )
	(if (or (equal planguage "c") (equal planguage "C"))
		(progn
		  (if comment (insert-string "  //interface to (X)Emacs compile facilities\n"))
		  (if (eq xref-platform 'windows)
			  (progn
				(insert-string "  -set compileproject \"make -f %s\\Makefile\"\n")
				)
			(insert-string "  -set compileproject \"make -f %s/Makefile\"\n")
		)))
	(if (or (equal planguage "j") (equal planguage "J")
			(equal planguage "b") (equal planguage "B"))
		(progn
		  (if (eq xref-platform 'windows)
			  (insert-string "  // -classpath c:\\java    ")
			(insert-string "  // -classpath /home/toto/java    ")
			)
		  (insert-string " // you may override CLASSPATH var\n")
		  (insert-string
"  // The path and packages with available JavaDoc documentation
  -htmljavadocpath=http://java.sun.com/j2se/1.3/docs/api
  -htmljavadocavailable=java.applet:java.awt:java.beans:java.io:java.lang:java.math:java.net:java.rmi:java.security:java.sql:java.text:java.util:javax.accessibility:javax.swing:org.omg.CORBA:org.omg.CosNaming
")
		  (if comment
			  (insert-string "  // few functions interfacing (X)Emacs compile facilities\n")
			)
		  (insert-string "  -set compilefile \"javac %s\"\n")
		  (if (eq xref-platform 'windows)
			  (progn
				(insert-string "  -set compiledir \"javac %s\\*.java\"\n")
				(insert-string "  -set compileproject \"javac -Xdepend %s\\Main.java\"\n")
				)
			(insert-string "  -set compiledir \"javac %s/*.java\"\n")
			(insert-string "  -set compileproject \"javac -Xdepend %s/Main.java\"\n")
			)
		  (insert-string "  -set run \"java com.toto.Main\"\n")
		  ))
))


(defun xref-project-new ()
" Create new project in  the .xrefrc Xrefactory option file. This puts
there  some common  options  for further  manual  edit. This  function
creates  only a  very simple  project which  can cover  small projects
stored in  a single  directory.  It is  usefull rather  for Xrefactory
beginners, advanced users will prefer to create and edit their .xrefrc
file manually.  See also  '.xrefrc' and 'xref'  manual pages  for more
info on format of .xrefrc file and available options.
"
  (interactive "")
  (let ((pname) (checked) (tname) (breakcheck) (planguage) (pcomments)
		(pact) (pedit) (sfiles))
	;; this is to create initial file (if not yet done)
	(call-completion 'assync 1 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	;;
	(setq pname (backslashify-name (file-directory-name (buffer-file-name))))
	(setq sfiles pname)
	(setq pname (read-from-minibuffer "Enter new project name :" pname))
	(find-file xref-options-file)
	(setq checked nil)
	(while (not checked)
	  (setq pname (remove-pending-slash pname))
	  (beginning-of-buffer)
	  (if (search-forward (concat "[" pname "]") nil t)
		  (setq pname (read-from-minibuffer 
					   "** Name already used, new name :" pname))
		(beginning-of-buffer)
		(if (or (search-forward (concat "[" pname "/") nil t)
				(search-forward (concat "[" pname "\\\\") nil t))
			(setq pname (read-from-minibuffer 
						 "** Name overlaps an existing, enter new name :" pname))
		  (setq tname pname)
		  (setq breakcheck nil)
		  (while (and tname (not breakcheck))
			(setq tname (file-directory-name tname))
			(beginning-of-buffer)
			(if (and tname 
					 (search-forward (concat "[" tname "]") nil t))
				(progn
				  (setq pname (read-from-minibuffer 
						 "** This name is yet covered, enter new name :" pname))
				  (setq breakcheck t)
				  )))
		  (if (not breakcheck)
			  (setq checked t)
			)
		)))
	;; The name is checked 
	(save-buffer) (kill-buffer nil)
	(setq planguage (read-from-minibuffer 
				 "This will be C or Java project (or Both) [cjb]? " "b"))
   	(setq pcomments (read-from-minibuffer 
				 "Do you wish that project description contains commentaries [yn]? " "y"))
   	(setq pfiles (read-from-minibuffer 
				 "Directory containing your sources? " sfiles))
	;; all questions done
	(find-file xref-options-file)
	(xref-append-new-project-section pname planguage pcomments pfiles)
	;;
	(save-buffer)
	(search-backward (concat "[" pname "]"))
	(recenter)

	(setq pact (read-from-minibuffer 
				 "Make the new project active [yn]? " "y"))
	(if (or (equal pact "y") (equal pact "Y"))
		(progn
		  (setq xref-current-project pname)
		  (message "Project '%s' is now active." pname)
		  )
	  )
	(setq pedit (read-from-minibuffer 
				 "Do you wish to adjust options manually [yn]? " "n"))
	(if (or (equal pedit "n") (equal pedit "N"))
		(kill-buffer nil)
	  )
	(message "Done.")
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-project-active () 
"Show currently  active project name. This function  is usefull mainly
if you  are not  sure which .xrefrc  section applies on  the currently
edited file.
"
  (interactive "")
  (if xref-current-project
	  (message "Active project is: %s" xref-current-project)
	(call-completion 'assync 1 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*) 
		(message "Auto detected project: %s" xref-last-message)
	  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-project-edit-options ()
"Edit manually .xrefrc file. This  function just load the .xrefrc file
and goes  to the active  project section. You  need to edit  this text
file manually. For more info  about .xrefrc file format, read 'xrefrc'
manual  page.  For  more  info about  available  options, read  'xref'
manual page.
"
  (interactive "")
  (let ((prjstr ""))
	;; get project name
	(call-completion 'assync 1 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*) 
		(setq prjstr (concat "[" xref-last-message "]"))
	  )
	(find-file xref-options-file)
	(beginning-of-buffer)
	(search-forward prjstr)
	(beginning-of-line)
	(next-line 1)
))


